"""Patching functions."""

from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Union

import numpy as np
from numpy.typing import NDArray

from ...utils.logging import get_logger
from ..dataset_utils import reshape_array
from ..dataset_utils.running_stats import compute_normalization_stats
from .sequential_patching import extract_patches_sequential

logger = get_logger(__name__)


@dataclass
class Stats:
    """Dataclass to store statistics."""

    means: Union[NDArray, tuple, list, None]
    """Mean of the data across channels."""

    stds: Union[NDArray, tuple, list, None]
    """Standard deviation of the data across channels."""

    def get_statistics(self) -> tuple[list[float], list[float]]:
        """Return the means and standard deviations.

        Returns
        -------
        tuple of two lists of floats
            Means and standard deviations.
        """
        if self.means is None or self.stds is None:
            return [], []

        return list(self.means), list(self.stds)


@dataclass
class PatchedOutput:
    """Dataclass to store patches and statistics."""

    patches: Union[NDArray]
    """Image patches."""

    targets: Union[NDArray, None]
    """Target patches."""

    image_stats: Stats
    """Statistics of the image patches."""

    target_stats: Stats
    """Statistics of the target patches."""


# called by in memory dataset
def prepare_patches_supervised(
    train_files: list[Path],
    target_files: list[Path],
    axes: str,
    patch_size: Union[list[int], tuple[int, ...]],
    read_source_func: Callable,
) -> PatchedOutput:
    """
    Iterate over data source and create an array of patches and corresponding targets.

    The lists of Paths should be pre-sorted.

    Parameters
    ----------
    train_files : list of pathlib.Path
        List of paths to training data.
    target_files : list of pathlib.Path
        List of paths to target data.
    axes : str
        Axes of the data.
    patch_size : list or tuple of int
        Size of the patches.
    read_source_func : Callable
        Function to read the data.

    Returns
    -------
    np.ndarray
        Array of patches.
    """
    means, stds, num_samples = 0, 0, 0
    all_patches, all_targets = [], []
    for train_filename, target_filename in zip(train_files, target_files):
        try:
            sample: np.ndarray = read_source_func(train_filename, axes)
            target: np.ndarray = read_source_func(target_filename, axes)
            means += sample.mean()
            stds += sample.std()
            num_samples += 1

            # reshape array
            sample = reshape_array(sample, axes)
            target = reshape_array(target, axes)

            # generate patches, return a generator
            patches, targets = extract_patches_sequential(
                sample, patch_size=patch_size, target=target
            )

            # convert generator to list and add to all_patches
            all_patches.append(patches)

            # ensure targets are not None (type checking)
            if targets is not None:
                all_targets.append(targets)
            else:
                raise ValueError(f"No target found for {target_filename}.")

        except Exception as e:
            # emit warning and continue
            logger.error(f"Failed to read {train_filename} or {target_filename}: {e}")

    # raise error if no valid samples found
    if num_samples == 0 or len(all_patches) == 0:
        raise ValueError(
            f"No valid samples found in the input data: {train_files} and "
            f"{target_files}."
        )

    image_means, image_stds = compute_normalization_stats(np.concatenate(all_patches))
    target_means, target_stds = compute_normalization_stats(np.concatenate(all_targets))

    patch_array: np.ndarray = np.concatenate(all_patches, axis=0)
    target_array: np.ndarray = np.concatenate(all_targets, axis=0)
    logger.info(f"Extracted {patch_array.shape[0]} patches from input array.")

    return PatchedOutput(
        patch_array,
        target_array,
        Stats(image_means, image_stds),
        Stats(target_means, target_stds),
    )


# called by in_memory_dataset
def prepare_patches_unsupervised(
    train_files: list[Path],
    axes: str,
    patch_size: Union[list[int], tuple[int]],
    read_source_func: Callable,
) -> PatchedOutput:
    """Iterate over data source and create an array of patches.

    This method returns the mean and standard deviation of the image.

    Parameters
    ----------
    train_files : list of pathlib.Path
        List of paths to training data.
    axes : str
        Axes of the data.
    patch_size : list or tuple of int
        Size of the patches.
    read_source_func : Callable
        Function to read the data.

    Returns
    -------
    PatchedOutput
        Dataclass holding patches and their statistics.
    """
    means, stds, num_samples = 0, 0, 0
    all_patches = []
    for filename in train_files:
        try:
            sample: np.ndarray = read_source_func(filename, axes)
            means += sample.mean()
            stds += sample.std()
            num_samples += 1

            # reshape array
            sample = reshape_array(sample, axes)

            # generate patches, return a generator
            patches, _ = extract_patches_sequential(sample, patch_size=patch_size)

            # convert generator to list and add to all_patches
            all_patches.append(patches)
        except Exception as e:
            # emit warning and continue
            logger.error(f"Failed to read {filename}: {e}")

    # raise error if no valid samples found
    if num_samples == 0:
        raise ValueError(f"No valid samples found in the input data: {train_files}.")

    image_means, image_stds = compute_normalization_stats(np.concatenate(all_patches))

    patch_array: np.ndarray = np.concatenate(all_patches)
    logger.info(f"Extracted {patch_array.shape[0]} patches from input array.")

    return PatchedOutput(
        patch_array, None, Stats(image_means, image_stds), Stats((), ())
    )


# called on arrays by in memory dataset
def prepare_patches_supervised_array(
    data: NDArray,
    axes: str,
    data_target: NDArray,
    patch_size: Union[list[int], tuple[int]],
) -> PatchedOutput:
    """Iterate over data source and create an array of patches.

    This method expects an array of shape SC(Z)YX, where S and C can be singleton
    dimensions.

    Patches returned are of shape SC(Z)YX, where S is now the patches dimension.

    Parameters
    ----------
    data : numpy.ndarray
        Input data array.
    axes : str
        Axes of the data.
    data_target : numpy.ndarray
        Target data array.
    patch_size : list or tuple of int
        Size of the patches.

    Returns
    -------
    PatchedOutput
        Dataclass holding the source and target patches, with their statistics.
    """
    # reshape array
    reshaped_sample = reshape_array(data, axes)
    reshaped_target = reshape_array(data_target, axes)

    # compute statistics
    image_means, image_stds = compute_normalization_stats(reshaped_sample)
    target_means, target_stds = compute_normalization_stats(reshaped_target)

    # generate patches, return a generator
    patches, patch_targets = extract_patches_sequential(
        reshaped_sample, patch_size=patch_size, target=reshaped_target
    )

    if patch_targets is None:
        raise ValueError("No target extracted.")

    logger.info(f"Extracted {patches.shape[0]} patches from input array.")

    return PatchedOutput(
        patches,
        patch_targets,
        Stats(image_means, image_stds),
        Stats(target_means, target_stds),
    )


# called by in memory dataset
def prepare_patches_unsupervised_array(
    data: NDArray,
    axes: str,
    patch_size: Union[list[int], tuple[int]],
) -> PatchedOutput:
    """
    Iterate over data source and create an array of patches.

    This method expects an array of shape SC(Z)YX, where S and C can be singleton
    dimensions.

    Patches returned are of shape SC(Z)YX, where S is now the patches dimension.

    Parameters
    ----------
    data : numpy.ndarray
        Input data array.
    axes : str
        Axes of the data.
    patch_size : list or tuple of int
        Size of the patches.

    Returns
    -------
    PatchedOutput
        Dataclass holding the patches and their statistics.
    """
    # reshape array
    reshaped_sample = reshape_array(data, axes)

    # calculate mean and std
    means, stds = compute_normalization_stats(reshaped_sample)

    # generate patches, return a generator
    patches, _ = extract_patches_sequential(reshaped_sample, patch_size=patch_size)

    return PatchedOutput(patches, None, Stats(means, stds), Stats((), ()))



import numpy as np
import pytest
import tifffile

from careamics.file_io.read import read_tiff


def test_read_tiff(tmp_path, ordered_array):
    """Test reading a tiff file."""
    # create an array
    array: np.ndarray = ordered_array((10, 10))

    # save files
    file = tmp_path / "test.tiff"
    tifffile.imwrite(file, array)

    # read files
    array_read = read_tiff(file)
    np.testing.assert_array_equal(array_read, array)


def test_read_tiff_invalid(tmp_path):
    # invalid file type
    file = tmp_path / "test.txt"
    file.write_text("test")
    with pytest.raises(ValueError):
        read_tiff(file)

    # non-existing file
    file = tmp_path / "test.tiff"
    with pytest.raises(FileNotFoundError):
        read_tiff(file)



"""Bioimage Model Zoo format functions."""

__all__ = [
    "create_model_description",
    "extract_model_path",
    "get_unzip_path",
    "create_env_text",
]

from .bioimage_utils import create_env_text, get_unzip_path
from .model_description import create_model_description, extract_model_path



import numpy as np
import pytest

from careamics.config.support import SupportedPixelManipulation
from careamics.transforms import N2VManipulate


@pytest.mark.parametrize(
    "strategy",
    [SupportedPixelManipulation.UNIFORM.value, SupportedPixelManipulation.MEDIAN.value],
)
def test_manipulate_n2v(strategy):
    """Test the N2V augmentation."""
    # create array, adding a channel to simulate a 2D image with channel last
    array = np.arange(16 * 16).reshape((16, 16))[np.newaxis, ...]

    # create augmentation
    aug = N2VManipulate(roi_size=5, masked_pixel_percentage=5, strategy=strategy)

    # apply augmentation
    augmented = aug(array)
    assert len(augmented) == 3  # transformed_patch, original_patch, mask

    # assert that the difference between the original and transformed patch are the
    # same pixels that are selected by the mask
    tr_path, orig_patch, mask = augmented
    diff_coords = np.array(np.where(tr_path != orig_patch))
    mask_coords = np.array(np.where(mask == 1))
    assert np.array_equal(diff_coords, mask_coords)



import numpy as np
from tqdm import tqdm


class EmptyPatchFetcher:
    """
    The idea is to fetch empty patches so that real content can be replaced with this.
    """

    def __init__(self, idx_manager, patch_size, data_frames, max_val_threshold=None):
        self._frames = data_frames
        self._idx_manager = idx_manager
        self._max_val_threshold = max_val_threshold
        self._idx_list = []
        self._patch_size = patch_size
        self._grid_size = 1
        self.set_empty_idx()

        print(f"[{self.__class__.__name__}] MaxVal:{self._max_val_threshold}")

    def compute_max(self, window):
        """
        Rolling compute.
        """
        N, H, W = self._frames.shape
        randnum = -954321
        assert self._grid_size == 1
        max_data = np.zeros((N, H - window, W - window)) * randnum

        for h in tqdm(range(H - window)):
            for w in range(W - window):
                max_data[:, h, w] = self._frames[:, h : h + window, w : w + window].max(
                    axis=(1, 2)
                )

        assert (max_data != 954321).any()
        return max_data

    def set_empty_idx(self):
        max_data = self.compute_max(self._patch_size)
        empty_loc = np.where(
            np.logical_and(max_data >= 0, max_data < self._max_val_threshold)
        )
        # print(max_data.shape, len(empty_loc))
        self._idx_list = []
        for idx in range(len(empty_loc[0])):
            n_idx = empty_loc[0][idx]
            h_start = empty_loc[1][idx]
            w_start = empty_loc[2][idx]
            # print(n_idx,h_start,w_start)
            # channel_idx = self._idx_manager.get_location_from_dataset_idx(0)[-1]
            loc = (n_idx, h_start, w_start, 0)
            idx = self._idx_manager.get_dataset_idx_from_location(loc)
            t, h, w, _ = self._idx_manager.get_location_from_dataset_idx(idx)
            assert h == h_start, f"{h} != {h_start}"
            assert w == w_start, f"{w} != {w_start}"
            assert t == n_idx, f"{t} != {n_idx}"
            self._idx_list.append(idx)

        self._idx_list = np.array(self._idx_list)

        assert len(self._idx_list) > 0

    def sample(self):
        return (np.random.choice(self._idx_list), self._grid_size)



from careamics.config.support import SupportedTransform
from careamics.transforms import get_all_transforms


def test_supported_transforms_in_accepted_transforms():
    """Test that all the supported transforms are in the accepted transforms."""
    for transform in SupportedTransform:
        assert transform in get_all_transforms()



import pytest

from careamics.config.architectures import UNetModel
from careamics.config.support import SupportedActivation


def test_instantiation():
    """Test that UNetModel can be instantiated."""
    model_params = {
        "architecture": "UNet",
        "conv_dim": 2,
        "num_channels_init": 16,
    }

    # instantiate model
    UNetModel(**model_params)


def test_architecture_missing():
    """Test that UNetModel requires architecture."""
    model_params = {
        "depth": 2,
        "num_channels_init": 16,
    }

    with pytest.raises(ValueError):
        UNetModel(**model_params)


@pytest.mark.parametrize("num_channels_init", [8, 16, 32, 96, 128])
def test_num_channels_init(num_channels_init: int):
    """Test that UNetModel accepts num_channels_init as an even number and
    minimum 8."""
    model_params = {"architecture": "UNet", "num_channels_init": num_channels_init}

    # instantiate model
    UNetModel(**model_params)


@pytest.mark.parametrize("num_channels_init", [2, 17, 127])
def test_wrong_num_channels_init(num_channels_init: int):
    """Test that wrong num_channels_init causes an error."""
    model_params = {"architecture": "UNet", "num_channels_init": num_channels_init}

    with pytest.raises(ValueError):
        UNetModel(**model_params)


def test_activations():
    """Test that UNetModel accepts all activations."""
    for act in SupportedActivation:
        if act != SupportedActivation.ELU:
            model_params = {
                "architecture": "UNet",
                "num_channels_init": 16,
                "final_activation": act.value,
            }

        # instantiate model
        UNetModel(**model_params)


def test_all_activations_are_supported():
    """Test that all activations defined in the Literal are supported."""
    # list of supported activations
    activations = list(SupportedActivation)

    # Algorithm json schema
    schema = UNetModel.model_json_schema()

    # check that all activations are supported
    for act in schema["properties"]["final_activation"]["enum"]:
        assert act in activations


def test_activation_wrong_values():
    """Test that wrong values are not accepted."""
    model_params = {
        "architecture": "UNet",
        "num_channels_init": 16,
        "final_activation": "wrong",
    }

    with pytest.raises(ValueError):
        UNetModel(**model_params)


def test_parameters_wrong_values_by_assigment():
    """Test that wrong values are not accepted through assignment."""
    model_params = {"architecture": "UNet", "num_channels_init": 16, "depth": 2}
    model = UNetModel(**model_params)

    # depth
    model.depth = model_params["depth"]
    with pytest.raises(ValueError):
        model.depth = -1

    # number of channels
    model.num_channels_init = model_params["num_channels_init"]
    with pytest.raises(ValueError):
        model.num_channels_init = 2


def test_model_dump():
    """Test that default values are excluded from model dump."""
    model_params = {
        "architecture": "UNet",
        "num_channels_init": 16,  # non-default value
        "final_activation": "ReLU",  # non-default value
    }
    model = UNetModel(**model_params)

    # dump model
    model_dict = model.model_dump(exclude_defaults=True)

    # check that default values are excluded except the architecture
    assert "architecture" not in model_dict
    assert len(model_dict) == 2

    # check that we get all the optional values with the exclude_defaults flag
    model_dict = model.model_dump(exclude_defaults=False)
    assert len(model_dict) == len(dict(model)) - 1



"""Custom model registration utilities."""

from typing import Callable

from torch.nn import Module

CUSTOM_MODELS = {}  # dictionary of custom models {"name": __class__}


def register_model(name: str) -> Callable:
    """Decorator used to register a torch.nn.Module class with a given `name`.

    Parameters
    ----------
    name : str
        Name of the model.

    Returns
    -------
    Callable
        Function allowing to instantiate the wrapped Module class.

    Raises
    ------
    ValueError
        If a model is already registered with that name.

    Examples
    --------
    ```python
    @register_model(name="linear")
    class LinearModel(nn.Module):
        def __init__(self, in_features, out_features):
            super().__init__()

            self.weight = nn.Parameter(ones(in_features, out_features))
            self.bias = nn.Parameter(ones(out_features))

        def forward(self, input):
            return (input @ self.weight) + self.bias
    ```
    """
    if name is None or name == "":
        raise ValueError("Model name cannot be empty.")

    if name in CUSTOM_MODELS:
        raise ValueError(
            f"Model {name} already exists. Choose a different name or run "
            f"`clear_custom_models()` to empty the registry."
        )

    def add_custom_model(model: Module) -> Module:
        """Add a custom model to the registry and return it.

        Parameters
        ----------
        model : Module
            Module class to register.

        Returns
        -------
        Module
            The registered model.
        """
        # add model to the registry
        CUSTOM_MODELS[name] = model

        return model

    return add_custom_model


def get_custom_model(name: str) -> Module:
    """Get the custom model corresponding to `name` from the registry.

    Parameters
    ----------
    name : str
        Name of the model to retrieve.

    Returns
    -------
    Module
        The requested model.

    Raises
    ------
    ValueError
        If the model is not registered.
    """
    if name not in CUSTOM_MODELS:
        raise ValueError(
            f"Model {name} is unknown. Have you registered it using "
            f'@register_model("{name}") as decorator?'
        )

    return CUSTOM_MODELS[name]


def clear_custom_models() -> None:
    """Clear the custom models registry."""
    # clear dictionary
    CUSTOM_MODELS.clear()



import pytest
import torch

from careamics.models.lvae.utils import crop_img_tensor, pad_img_tensor


@pytest.mark.parametrize(
    "x, size",
    [
        (torch.zeros((1, 1, 64, 64)), (32, 32)),
        (torch.zeros((1, 1, 64, 64, 64)), (32, 32, 32)),
    ],
)
def test_crop_img_shape(x, size):
    res = crop_img_tensor(x, size)
    assert res.shape[2:] == size


@pytest.mark.parametrize(
    "x, size",
    [
        (torch.zeros((1, 1, 64, 64)), (65, 32)),
        (torch.zeros((1, 1, 64, 64, 64)), (65, 32, 32)),
    ],
)
def test_crop_img_assert(x, size):
    with pytest.raises(ValueError):
        crop_img_tensor(x, size)


@pytest.mark.parametrize(
    "x, size, expected",
    [
        (
            torch.arange(1, 17).reshape((1, 1, 4, 4)),
            (2, 2),
            torch.tensor([[[[6, 7], [10, 11]]]]),
        ),
        (
            torch.arange(1, 28).reshape((1, 1, 3, 3, 3)),
            (2, 2, 2),
            torch.tensor([[[[[1, 2], [4, 5]], [[10, 11], [13, 14]]]]]),
        ),
    ],
)
def test_crop_img_result(x, size, expected):
    res = crop_img_tensor(x, size)
    assert torch.equal(res, expected)


@pytest.mark.parametrize(
    "x, size",
    [
        (torch.zeros((1, 1, 64, 64)), (96, 96)),
        (torch.zeros((1, 1, 64, 64, 64)), (96, 96, 96)),
    ],
)
def test_pad_img_shape(x, size):
    res = pad_img_tensor(x, size)
    assert res.shape[2:] == size


@pytest.mark.parametrize(
    "x, size",
    [
        (torch.zeros((1, 1, 64, 64)), (65, 32)),
        (torch.zeros((1, 1, 64, 64, 64)), (65, 32, 32)),
    ],
)
def test_pad_img_assert(x, size):
    with pytest.raises(ValueError):
        pad_img_tensor(x, size)


@pytest.mark.parametrize(
    "x, size, expected",
    [
        (
            torch.arange(1, 10).reshape((1, 1, 3, 3)),
            (5, 5),
            torch.tensor(
                [
                    [
                        [
                            [0, 0, 0, 0, 0],
                            [0, 1, 2, 3, 0],
                            [0, 4, 5, 6, 0],
                            [0, 7, 8, 9, 0],
                            [0, 0, 0, 0, 0],
                        ]
                    ]
                ]
            ),
        ),
        (
            torch.arange(1, 9).reshape((1, 1, 2, 2, 2)),
            (4, 4, 4),
            torch.tensor(
                [
                    [
                        [
                            [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]],
                            [[0, 0, 0, 0], [0, 1, 2, 0], [0, 3, 4, 0], [0, 0, 0, 0]],
                            [[0, 0, 0, 0], [0, 5, 6, 0], [0, 7, 8, 0], [0, 0, 0, 0]],
                            [[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]],
                        ]
                    ]
                ]
            ),
        ),
    ],
)
def test_pad_img_result(x, size, expected):
    res = pad_img_tensor(x, size)
    assert torch.equal(res, expected)



import pytest

from careamics.config.support import SupportedData
from careamics.file_io import get_read_func
from careamics.file_io.read import read_tiff


def test_get_read_tiff():
    assert get_read_func(SupportedData.TIFF) is read_tiff


def test_get_read_any_error():
    with pytest.raises(NotImplementedError):
        get_read_func("some random")



import numpy as np
import pytest
import tifffile

from careamics.config import DataConfig
from careamics.config.support import SupportedData
from careamics.dataset import InMemoryDataset


def test_number_of_patches(ordered_array):
    """Test the number of patches extracted from InMemoryDataset."""
    # create array
    array = ordered_array((20, 20))

    # create config
    config_dict = {
        "data_type": SupportedData.ARRAY.value,
        "patch_size": [8, 8],
        "axes": "YX",
    }
    config = DataConfig(**config_dict)

    # create dataset
    dataset = InMemoryDataset(
        data_config=config,
        inputs=array,
    )

    # check number of patches
    assert len(dataset) == dataset.data.shape[0]


@pytest.mark.parametrize("percentage", [0.1, 0.6])
def test_extracting_val_array(ordered_array, percentage):
    """Test extracting a validation set patches from InMemoryDataset."""
    # create array
    array = ordered_array((32, 32))

    # create config
    config_dict = {
        "data_type": SupportedData.ARRAY.value,
        "patch_size": [8, 8],
        "axes": "YX",
    }
    config = DataConfig(**config_dict)

    # create dataset
    dataset = InMemoryDataset(
        data_config=config,
        inputs=array,
    )

    # compute number of patches
    total_n_patches = len(dataset)
    minimum_patches = 5
    n_patches = max(round(percentage * total_n_patches), minimum_patches)

    # extract datset
    valset = dataset.split_dataset(percentage, minimum_patches)

    # check number of patches
    assert len(valset) == n_patches
    assert len(dataset) == total_n_patches - n_patches

    # check that none of the validation patch values are in the original dataset
    assert np.in1d(valset.data, dataset.data).sum() == 0


@pytest.mark.parametrize("percentage", [0.1, 0.6])
def test_extracting_val_files(tmp_path, ordered_array, percentage):
    """Test extracting a validation set patches from InMemoryDataset."""
    # create array
    array = ordered_array((32, 32))

    # save array to file
    file_path = tmp_path / "array.tif"
    tifffile.imwrite(file_path, array)

    # create config
    config_dict = {
        "data_type": SupportedData.ARRAY.value,
        "patch_size": [8, 8],
        "axes": "YX",
    }
    config = DataConfig(**config_dict)

    # create dataset
    dataset = InMemoryDataset(
        data_config=config,
        inputs=[file_path],
    )

    # compute number of patches
    total_n_patches = len(dataset)
    minimum_patches = 5
    n_patches = max(round(percentage * total_n_patches), minimum_patches)

    # extract datset
    valset = dataset.split_dataset(percentage, minimum_patches)

    # check number of patches
    assert len(valset) == n_patches
    assert len(dataset) == total_n_patches - n_patches

    # check that none of the validation patch values are in the original dataset
    assert np.in1d(valset.data, dataset.data).sum() == 0



from typing import Union

import numpy as np
import torch
from scipy import stats


def get_last_index(bin_count, quantile):
    cumsum = np.cumsum(bin_count)
    normalized_cumsum = cumsum / cumsum[-1]
    for i in range(1, len(normalized_cumsum)):
        if normalized_cumsum[-i] < quantile:
            return i - 1
    return None


def get_first_index(bin_count, quantile):
    cumsum = np.cumsum(bin_count)
    normalized_cumsum = cumsum / cumsum[-1]
    for i in range(len(normalized_cumsum)):
        if normalized_cumsum[i] > quantile:
            return i
    return None


class Calibration:
    """Calibrate the uncertainty computed over samples from LVAE model.

    Calibration is done by learning a scalar that maps the pixel-wise standard
    deviation of the the predicted samples into the actual prediction error.
    """

    def __init__(self, num_bins: int = 15):
        self._bins = num_bins
        self._bin_boundaries = None

    def logvar_to_std(self, logvar: np.ndarray) -> np.ndarray:
        return np.exp(logvar / 2)

    def compute_bin_boundaries(self, predict_std: np.ndarray) -> np.ndarray:
        """Compute the bin boundaries for `num_bins` bins and predicted std values."""
        min_std = np.min(predict_std)
        max_std = np.max(predict_std)
        return np.linspace(min_std, max_std, self._bins + 1)

    def compute_stats(
        self, pred: np.ndarray, pred_std: np.ndarray, target: np.ndarray
    ) -> dict[int, dict[str, Union[np.ndarray, list]]]:
        """
        It computes the bin-wise RMSE and RMV for each channel of the predicted image.

        Recall that:
            - RMSE = np.sqrt((pred - target)**2 / num_pixels)
            - RMV = np.sqrt(np.mean(pred_std**2))

        ALGORITHM
        - For each channel:
            - Given the bin boundaries, assign pixels of `std_ch` array to a specific bin index.
            - For each bin index:
                - Compute the RMSE, RMV, and number of pixels for that bin.

        NOTE: each channel of the predicted image/logvar has its own stats.

        Parameters
        ----------
        pred: np.ndarray
            Predicted patches, shape (n, h, w, c).
        pred_std: np.ndarray
            Std computed over the predicted patches, shape (n, h, w, c).
        target: np.ndarray
            Target GT image, shape (n, h, w, c).
        """
        self._bin_boundaries = {}
        stats_dict = {}
        for ch_idx in range(pred.shape[-1]):
            stats_dict[ch_idx] = {
                "bin_count": [],
                "rmv": [],
                "rmse": [],
                "bin_boundaries": None,
                "bin_matrix": [],
                "rmse_err": [],
            }
            pred_ch = pred[..., ch_idx]
            std_ch = pred_std[..., ch_idx]
            target_ch = target[..., ch_idx]
            boundaries = self.compute_bin_boundaries(std_ch)
            stats_dict[ch_idx]["bin_boundaries"] = boundaries
            bin_matrix = np.digitize(std_ch.reshape(-1), boundaries)
            bin_matrix = bin_matrix.reshape(std_ch.shape)
            stats_dict[ch_idx]["bin_matrix"] = bin_matrix
            error = (pred_ch - target_ch) ** 2
            for bin_idx in range(1, 1 + self._bins):
                bin_mask = bin_matrix == bin_idx
                bin_error = error[bin_mask]
                bin_size = np.sum(bin_mask)
                bin_error = (
                    np.sqrt(np.sum(bin_error) / bin_size) if bin_size > 0 else None
                )
                stderr = (
                    np.std(error[bin_mask]) / np.sqrt(bin_size)
                    if bin_size > 0
                    else None
                )
                rmse_stderr = np.sqrt(stderr) if stderr is not None else None

                bin_var = np.mean((std_ch[bin_mask] ** 2))
                stats_dict[ch_idx]["rmse"].append(bin_error)
                stats_dict[ch_idx]["rmse_err"].append(rmse_stderr)
                stats_dict[ch_idx]["rmv"].append(np.sqrt(bin_var))
                stats_dict[ch_idx]["bin_count"].append(bin_size)
        return stats_dict


def get_calibrated_factor_for_stdev(
    pred: Union[np.ndarray, torch.Tensor],
    pred_std: Union[np.ndarray, torch.Tensor],
    target: Union[np.ndarray, torch.Tensor],
    q_s: float = 0.00001,
    q_e: float = 0.99999,
    num_bins: int = 30,
) -> dict[str, float]:
    """Calibrate the uncertainty by multiplying the predicted std with a scalar.

    Parameters
    ----------
    pred : Union[np.ndarray, torch.Tensor]
        Predicted image, shape (n, h, w, c).
    pred_std : Union[np.ndarray, torch.Tensor]
        Predicted std, shape (n, h, w, c).
    target : Union[np.ndarray, torch.Tensor]
        Target image, shape (n, h, w, c).
    q_s : float, optional
        Start quantile, by default 0.00001.
    q_e : float, optional
        End quantile, by default 0.99999.
    num_bins : int, optional
        Number of bins to use for calibration, by default 30.

    Returns
    -------
    dict[str, float]
        Calibrated factor for each channel (slope + intercept).
    """
    calib = Calibration(num_bins=num_bins)
    stats_dict = calib.compute_stats(pred, pred_std, target)
    outputs = {}
    for ch_idx in stats_dict.keys():
        y = stats_dict[ch_idx]["rmse"]
        x = stats_dict[ch_idx]["rmv"]
        count = stats_dict[ch_idx]["bin_count"]

        first_idx = get_first_index(count, q_s)
        last_idx = get_last_index(count, q_e)
        x = x[first_idx:-last_idx]
        y = y[first_idx:-last_idx]
        slope, intercept, *_ = stats.linregress(x, y)
        output = {"scalar": slope, "offset": intercept}
        outputs[ch_idx] = output
    return outputs


def plot_calibration(ax, calibration_stats):
    first_idx = get_first_index(calibration_stats[0]["bin_count"], 0.001)
    last_idx = get_last_index(calibration_stats[0]["bin_count"], 0.999)
    ax.plot(
        calibration_stats[0]["rmv"][first_idx:-last_idx],
        calibration_stats[0]["rmse"][first_idx:-last_idx],
        "o",
        label=r"$\hat{C}_0$: Ch1",
    )

    first_idx = get_first_index(calibration_stats[1]["bin_count"], 0.001)
    last_idx = get_last_index(calibration_stats[1]["bin_count"], 0.999)
    ax.plot(
        calibration_stats[1]["rmv"][first_idx:-last_idx],
        calibration_stats[1]["rmse"][first_idx:-last_idx],
        "o",
        label=r"$\hat{C}_1: : Ch2$",
    )

    ax.set_xlabel("RMV")
    ax.set_ylabel("RMSE")
    ax.legend()



#!/usr/bin/env python
"""Export a README file for the bioimage model zoo."""
from pathlib import Path

from careamics.config import create_n2v_configuration
from careamics.model_io.bioimage._readme_factory import readme_factory


def main():
    # create configuration
    config = create_n2v_configuration(
        experiment_name="export_bmz_readme",
        data_type="array",
        axes="YX",
        patch_size=(64, 64),
        batch_size=2,
        num_epochs=10,
    )
    # export README
    readme_path = readme_factory(
        config=config, careamics_version="0.1.0", data_description="Mydata"
    )

    # copy file to __file__
    readme_path.rename(Path(__file__).parent / "README.md")


if __name__ == "__main__":
    main()



import numpy as np
import pytest

from careamics.dataset.dataset_utils.running_stats import compute_normalization_stats


@pytest.mark.parametrize("samples, channels", [[1, 2], [1, 2]])
def test_compute_normalization_stats(samples, channels):
    """Test the compute_normalization_stats function."""
    # Create data
    array = np.arange(100 * samples * channels).reshape((samples, channels, 10, 10))

    # Compute mean and std
    mean, std = compute_normalization_stats(image=array)
    for ch in range(array.shape[1]):
        assert np.isclose(mean[ch], array[:, ch, ...].mean())
        assert np.isclose(std[ch], array[:, ch, ...].std())

    # Create data 3D
    array = np.arange(1000 * samples * channels).reshape(
        (samples, channels, 10, 10, 10)
    )

    # Compute mean and std
    mean, std = compute_normalization_stats(image=array)
    for ch in range(array.shape[1]):
        assert np.isclose(mean[ch], array[:, ch, ...].mean())
        assert np.isclose(std[ch], array[:, ch, ...].std())



"""Test PredictionWriterCallback class."""

import os
from pathlib import Path
from typing import Union
from unittest.mock import Mock, patch

import numpy as np
import pytest
import tifffile
from pytorch_lightning import LightningModule, Trainer
from pytorch_lightning.callbacks import ModelCheckpoint
from torch.utils.data import DataLoader

from careamics.config import Configuration
from careamics.config.support import SupportedData
from careamics.dataset import IterablePredDataset
from careamics.lightning import (
    create_careamics_module,
    create_predict_datamodule,
    create_train_datamodule,
)
from careamics.lightning.callbacks import PredictionWriterCallback
from careamics.lightning.callbacks.prediction_writer_callback import (
    WriteStrategy,
    create_write_strategy,
)
from careamics.prediction_utils import convert_outputs


@pytest.fixture
def write_strategy():
    """Mock `WriteFunc`."""
    return Mock(spec=WriteStrategy)


@pytest.fixture
def dirpath(tmp_path: Path):
    """Directory path."""
    return tmp_path / "predictions"


@pytest.fixture
def prediction_writer_callback(
    write_strategy: WriteStrategy, dirpath: Union[Path, str]
):
    """Initialized `PredictionWriterCallback`."""
    return PredictionWriterCallback(write_strategy=write_strategy, dirpath=dirpath)


# TODO: smoke test with tiff (& example custom save func?)


def test_smoke_n2v_tiled_tiff(tmp_path, minimum_configuration):
    rng = np.random.default_rng(42)

    # training data
    train_array = rng.integers(0, 255, (32, 32)).astype(np.float32)
    val_array = rng.integers(0, 255, (32, 32)).astype(np.float32)

    # write train array to tiff
    train_dir = tmp_path / "train"
    train_dir.mkdir()
    file_name = "image.tiff"
    train_file = train_dir / file_name
    tifffile.imwrite(train_file, train_array)

    cfg = Configuration(**minimum_configuration)

    # create lightning module
    model = create_careamics_module(
        algorithm=cfg.algorithm_config.algorithm,
        loss=cfg.algorithm_config.loss,
        architecture=cfg.algorithm_config.model.architecture,
    )

    # create data module
    data = create_train_datamodule(
        train_data=train_array,
        val_data=val_array,
        data_type=cfg.data_config.data_type,
        patch_size=cfg.data_config.patch_size,
        axes=cfg.data_config.axes,
        batch_size=cfg.data_config.batch_size,
    )

    # create prediction writer callback params
    write_strategy = create_write_strategy(write_type="tiff", tiled=True)
    dirpath = tmp_path / "predictions"

    # create trainer
    trainer = Trainer(
        max_epochs=1,
        default_root_dir=tmp_path,
        callbacks=[
            ModelCheckpoint(
                dirpath=tmp_path / "checkpoints",
                filename="test_lightning_api",
            ),
            PredictionWriterCallback(write_strategy=write_strategy, dirpath=dirpath),
        ],
    )

    # train
    trainer.fit(model, datamodule=data)

    # predict
    means, stds = data.get_data_statistics()
    predict_data = create_predict_datamodule(
        data_type="tiff",
        pred_data=train_dir,
        axes=cfg.data_config.axes,
        image_means=means,
        image_stds=stds,
        tile_size=(8, 8),
        tile_overlap=(2, 2),
    )

    # predict
    predicted = trainer.predict(model, datamodule=predict_data)
    predicted_images = convert_outputs(predicted, tiled=True)

    # assert predicted file exists
    assert (dirpath / file_name).is_file()

    # open file
    save_data = tifffile.imread(dirpath / file_name)
    # save data has singleton channel axis
    np.testing.assert_array_equal(save_data, predicted_images[0][0], verbose=True)


def test_smoke_n2v_untiled_tiff(tmp_path, minimum_configuration):
    rng = np.random.default_rng(42)

    # training data
    train_array = rng.integers(0, 255, (32, 32)).astype(np.float32)
    val_array = rng.integers(0, 255, (32, 32)).astype(np.float32)

    # write train array to tiff
    train_dir = tmp_path / "train"
    train_dir.mkdir()
    file_name = "image.tiff"
    train_file = train_dir / file_name
    tifffile.imwrite(train_file, train_array)

    cfg = Configuration(**minimum_configuration)

    # create lightning module
    model = create_careamics_module(
        algorithm=cfg.algorithm_config.algorithm,
        loss=cfg.algorithm_config.loss,
        architecture=cfg.algorithm_config.model.architecture,
    )

    # create data module
    data = create_train_datamodule(
        train_data=train_array,
        val_data=val_array,
        data_type=cfg.data_config.data_type,
        patch_size=cfg.data_config.patch_size,
        axes=cfg.data_config.axes,
        batch_size=cfg.data_config.batch_size,
    )

    # create prediction writer callback params
    write_strategy = create_write_strategy(write_type="tiff", tiled=False)
    dirpath = tmp_path / "predictions"

    # create trainer
    trainer = Trainer(
        max_epochs=1,
        default_root_dir=tmp_path,
        callbacks=[
            ModelCheckpoint(
                dirpath=tmp_path / "checkpoints",
                filename="test_lightning_api",
            ),
            PredictionWriterCallback(write_strategy=write_strategy, dirpath=dirpath),
        ],
    )

    # train
    trainer.fit(model, datamodule=data)

    # predict
    means, stds = data.get_data_statistics()
    predict_data = create_predict_datamodule(
        data_type="tiff",
        pred_data=train_dir,
        axes=cfg.data_config.axes,
        image_means=means,
        image_stds=stds,
    )

    # predict
    predicted = trainer.predict(model, datamodule=predict_data)
    predicted_images = convert_outputs(predicted, tiled=False)

    # assert predicted file exists
    assert (dirpath / file_name).is_file()

    # open file
    save_data = tifffile.imread(dirpath / file_name)
    # save data has singleton channel axis
    np.testing.assert_array_equal(save_data, predicted_images[0][0], verbose=True)


def test_initialization(prediction_writer_callback, write_strategy, dirpath):
    """Test `PredictionWriterCallback` initializes as expected."""
    assert prediction_writer_callback.writing_predictions is True
    assert prediction_writer_callback.write_strategy is write_strategy
    assert prediction_writer_callback.dirpath == Path(dirpath).resolve()


def test_init_dirpath_absolute_path(prediction_writer_callback):
    """Test initialization of dirpath with absolute path."""
    absolute_path = Path("/absolute/path").absolute()
    prediction_writer_callback._init_dirpath(absolute_path)
    assert prediction_writer_callback.dirpath == absolute_path


def test_init_dirpath_relative_path(prediction_writer_callback):
    """Test initialization of dirpath with relatice path."""
    relative_path = "relative/path"
    # patch pathlib.Path.cwd to return
    mock_cwd = Path("/current/working/dir")
    with patch("pathlib.Path.cwd", return_value=mock_cwd):
        prediction_writer_callback._init_dirpath(relative_path)
        assert prediction_writer_callback.dirpath == mock_cwd / relative_path


def test_setup_prediction_directory_creation(prediction_writer_callback, dirpath):
    """
    Test prediction directory is created when `setup` is called at `stage="predict"`.
    """
    trainer = Mock(spec=Trainer)
    pl_module = Mock(spec=LightningModule)
    stage = "predict"

    prediction_writer_callback.setup(trainer=trainer, pl_module=pl_module, stage=stage)

    assert os.path.isdir(dirpath)


def test_write_on_batch_end(prediction_writer_callback):
    """
    Test `PredictionWriterCallback.write_on_batch_end`.

    Check `prediction_writer_callback.write_strategy.write_batch` is called as
    expected.
    """
    # mock write_on_batch_end inputs.
    trainer = Mock(spec=Trainer)
    pl_module = Mock(spec=LightningModule)
    prediction = Mock()
    batch_indices = [0, 1, 2]
    batch = Mock()
    batch_idx = 0
    dataloader_idx = 0

    # Mocking the dataset to be of type IterablePredDataset
    mock_dataset = Mock(spec=IterablePredDataset)
    trainer.predict_dataloaders = [Mock(spec=DataLoader)]
    trainer.predict_dataloaders[dataloader_idx].dataset = mock_dataset

    prediction_writer_callback.write_on_batch_end(
        trainer, pl_module, prediction, batch_indices, batch, batch_idx, dataloader_idx
    )
    # The different strategy functions are tested individually
    #   so they do not need to be tested here
    prediction_writer_callback.write_strategy.write_batch.assert_called_once_with(
        trainer=trainer,
        pl_module=pl_module,
        prediction=prediction,
        batch_indices=batch_indices,
        batch=batch,
        batch_idx=batch_idx,
        dataloader_idx=dataloader_idx,
        dirpath=prediction_writer_callback.dirpath,
    )


def test_write_on_batch_end_writing_predictions_off(prediction_writer_callback):
    """
    Test that `write_batch` acts as expected when writing predictions is set to off.

    Ensure `PredictionWriterCallback.write_strategy.write_batch` is not called when
    `PredictionWriterCallback.writing_predictions=False`.
    """
    prediction_writer_callback.writing_predictions = False

    trainer = Mock(spec=Trainer)
    pl_module = Mock(spec=LightningModule)
    prediction = Mock()
    batch_indices = [0, 1, 2]
    batch = Mock()
    batch_idx = 0
    dataloader_idx = 0

    # Mocking the dataset to be of type IterablePredDataset or IterableTiledPredDataset
    mock_dataset = Mock(spec=IterablePredDataset)
    trainer.predict_dataloaders = [Mock(spec=DataLoader)]
    trainer.predict_dataloaders[dataloader_idx].dataset = mock_dataset

    prediction_writer_callback.write_on_batch_end(
        trainer, pl_module, prediction, batch_indices, batch, batch_idx, dataloader_idx
    )
    prediction_writer_callback.write_strategy.write_batch.assert_not_called()


def test_from_write_func_params(write_strategy, dirpath):
    """
    Test `from_write_func_param` classmethod initializes `PredictWriterCallback`.
    """

    write_type = SupportedData.TIFF
    tiled = True
    write_func = None
    write_extension = None
    write_func_kwargs = None

    # mock create_write_strategy as it is already tested individually.
    with patch(
        "careamics.lightning.callbacks"
        ".prediction_writer_callback.prediction_writer_callback.create_write_strategy"
    ) as mock_create_write_strategy:
        write_strategy = Mock(spec=WriteStrategy)
        mock_create_write_strategy.return_value = write_strategy
        callback = PredictionWriterCallback.from_write_func_params(
            write_type=write_type,
            tiled=tiled,
            write_func=write_func,
            write_extension=write_extension,
            write_func_kwargs=write_func_kwargs,
            dirpath=dirpath,
        )
        mock_create_write_strategy.assert_called_once_with(
            write_type=write_type,
            tiled=tiled,
            write_func=write_func,
            write_extension=write_extension,
            write_func_kwargs=write_func_kwargs,
        )
        assert callback.write_strategy == write_strategy
        assert callback.dirpath == Path(dirpath).resolve()



"""Configuration module."""

__all__ = [
    "FCNAlgorithmConfig",
    "VAEAlgorithmConfig",
    "DataConfig",
    "Configuration",
    "CheckpointModel",
    "InferenceConfig",
    "load_configuration",
    "save_configuration",
    "TrainingConfig",
    "create_n2v_configuration",
    "create_n2n_configuration",
    "create_care_configuration",
    "register_model",
    "CustomModel",
    "clear_custom_models",
    "GaussianMixtureNMConfig",
    "MultiChannelNMConfig",
    "LVAELossConfig",
]
from .architectures import CustomModel, clear_custom_models, register_model
from .callback_model import CheckpointModel
from .configuration_factory import (
    create_care_configuration,
    create_n2n_configuration,
    create_n2v_configuration,
)
from .configuration_model import (
    Configuration,
    load_configuration,
    save_configuration,
)
from .data_model import DataConfig
from .fcn_algorithm_model import FCNAlgorithmConfig
from .inference_model import InferenceConfig
from .loss_model import LVAELossConfig
from .nm_model import GaussianMixtureNMConfig, MultiChannelNMConfig
from .training_model import TrainingConfig
from .vae_algorithm_model import VAEAlgorithmConfig



import numpy as np
import pytest

from careamics.transforms.pixel_manipulation import (
    _apply_struct_mask,
    _get_stratified_coords,
    median_manipulate,
    uniform_manipulate,
)
from careamics.transforms.struct_mask_parameters import StructMaskParameters


@pytest.mark.parametrize(
    "mask_pixel_perc, shape, num_iterations",
    [(0.4, (32, 32), 1000), (0.4, (10, 10, 10), 1000)],
)
def test_get_stratified_coords(mask_pixel_perc, shape, num_iterations):
    """Test the get_stratified_coords function.

    Ensure that the array of coordinates is randomly distributed across the
    image and that most pixels get selected.
    """
    rng = np.random.default_rng(42)

    # Define the dummy array
    array = np.zeros(shape)

    # Iterate over the number of iterations and add the coordinates. This is an MC
    # simulation to ensure that the coordinates are randomly distributed and not
    # biased towards any particular region.
    for _ in range(num_iterations):
        # Get the coordinates of the pixels to be masked
        coords = _get_stratified_coords(mask_pixel_perc, shape, rng)

        # Check that there is at least one coordinate chosen
        assert len(coords) > 0

        # Check every pair in the array of coordinates
        for coord_pair in coords:
            # Check that the coordinates are of the same shape as the patch dims
            assert len(coord_pair) == len(shape)

            # Check that the coordinates are positive values
            assert all(coord_pair) >= 0

            # Check that the coordinates are within the shape of the array
            assert [c <= s for c, s in zip(coord_pair, shape)]

        # Add the 1 to the every coordinate location.
        array[tuple(np.array(coords).T.tolist())] += 1

    # Ensure that there's no strong pattern in the array and sufficient number of
    # pixels is masked.
    assert np.sum(array == 0) < np.sum(shape)


@pytest.mark.parametrize("shape", [(8, 8), (3, 8, 8), (8, 8, 8)])
def test_uniform_manipulate(ordered_array, shape):
    """Test the uniform_manipulate function.

    Ensures that the mask corresponds to the manipulated pixels, and that the
    manipulated pixels have a value taken from a ROI surrounding them.
    """
    rng = np.random.default_rng(42)

    # create the array
    patch = ordered_array(shape)

    # manipulate the array
    transform_patch, mask = uniform_manipulate(
        patch, mask_pixel_percentage=10, subpatch_size=5, rng=rng
    )

    # find pixels that have different values between patch and transformed patch
    diff_coords = np.array(np.where(patch != transform_patch))

    # find non-zero pixels in the mask
    mask_coords = np.array(np.where(mask == 1))

    # check that the transformed pixels correspond to the masked pixels
    assert np.array_equal(diff_coords, mask_coords)

    # for each pixel masked, check that the manipulated pixel value is within the roi
    for i in range(mask_coords.shape[-1]):
        # get coordinates
        coords = mask_coords[..., i]

        # get roi using slice in each dimension
        slices = tuple(
            [
                slice(max(0, coords[i] - 2), min(shape[i], coords[i] + 3))
                for i in range(-coords.shape[0] + 1, 0)  # range -4, -3, -2, -1
            ]
        )
        roi = patch[
            (...,) + slices
        ]  # TODO ellipsis needed bc singleton dim, might need to go away

        # TODO needs to be revisited !
        # check that the pixel value comes from the actual roi
        assert transform_patch[tuple(coords)] in roi


@pytest.mark.parametrize("shape", [(8, 8), (3, 8, 8), (8, 8, 8)])
def test_median_manipulate(ordered_array, shape):
    """Test the uniform_manipulate function.

    Ensures that the mask corresponds to the manipulated pixels, and that the
    manipulated pixels have a value taken from a ROI surrounding them.
    """
    rng = np.random.default_rng(42)

    # create the array
    patch = ordered_array(shape).astype(np.float32)

    # manipulate the array
    transform_patch, mask = median_manipulate(
        patch, subpatch_size=5, mask_pixel_percentage=10, rng=rng
    )

    # find pixels that have different values between patch and transformed patch
    diff_coords = np.array(np.where(patch != transform_patch))

    # find non-zero pixels in the mask
    mask_coords = np.array(np.where(mask == 1))

    # check that the transformed pixels correspond to the masked pixels
    assert np.array_equal(diff_coords, mask_coords)

    # for each pixel masked, check that the manipulated pixel value is within the roi
    for i in range(mask_coords.shape[-1]):
        # get coordinates
        coords = mask_coords[..., i]

        # get roi using slice in each dimension
        slices = tuple(
            [
                slice(max(0, coords[i] - 2), min(shape[i], coords[i] + 3))
                for i in range(coords.shape[0])  # range -4, -3, -2, -1
            ]
        )
        roi = patch[tuple(slices)]

        # remove value of roi center from roi
        roi = roi[roi != patch[tuple(coords)]]

        # check that the pixel value comes from the actual roi
        assert transform_patch[tuple(coords)] == np.median(roi)


@pytest.mark.parametrize(
    "coords, struct_axis, struct_span",
    [((2, 2), 1, 5), ((3, 4), 0, 5), ((9, 0), 0, 5), (((1, 2), (3, 4)), 1, 5)],
)
def test_apply_struct_mask(coords, struct_axis, struct_span):
    """Test the uniform_manipulate function.

    Ensures that the mask corresponds to the manipulated pixels, and that the
    manipulated pixels have a value taken from a ROI surrounding them.
    """
    rng = np.random.default_rng(42)

    struct_params = StructMaskParameters(axis=struct_axis, span=struct_span)

    # create array
    patch = np.arange(
        100,
    ).reshape((10, 10))

    # make a copy of the original patch for comparison
    original_patch = patch.copy()
    coords = np.array(coords)

    # expand the coords if only one roi is given
    if coords.ndim == 1:
        coords = coords[None, :]

    # manipulate the array
    transform_patch = _apply_struct_mask(
        patch,
        coords=coords,
        struct_params=struct_params,
        rng=rng,
    )
    changed_values = patch[np.where(original_patch != transform_patch)]

    # check that the transformed pixels correspond to the masked pixels
    transformed = []
    axis = 1 - struct_axis
    for i in range(coords.shape[0]):
        # get indices to mask
        indices_to_mask = [
            c
            for c in range(
                max(0, coords[i, axis] - struct_span // 2),
                min(transform_patch.shape[1], coords[i, axis] + struct_span // 2) + 1,
            )
            if c != coords[i, axis]
        ]

        # add to transform
        if struct_axis == 0:
            transformed.append(transform_patch[coords[i, 0]][indices_to_mask])
        else:
            transformed.append(transform_patch[:, coords[i, 1]][indices_to_mask])

    assert np.array_equal(
        np.sort(changed_values), np.sort(np.concatenate(transformed, axis=0))
    )


@pytest.mark.parametrize(
    "coords, struct_axis, struct_span",
    [
        ((1, 2, 2), 1, 5),
        ((2, 3, 4), 0, 5),
        ((0, 9, 0), 0, 5),
        (((2, 1, 2), (1, 9, 0), (0, 3, 4)), 1, 5),
    ],
)
def test_apply_struct_mask_3D(coords, struct_axis, struct_span):
    """Test the uniform_manipulate function.

    Ensures that the mask corresponds to the manipulated pixels, and that the
    manipulated pixels have a value taken from a ROI surrounding them.
    """
    rng = np.random.default_rng(42)

    struct_params = StructMaskParameters(axis=struct_axis, span=struct_span)

    # create array
    patch = np.arange(
        100 * 3,
    ).reshape((3, 10, 10))

    # make a copy of the original patch for comparison
    original_patch = patch.copy()
    coords = np.array(coords)

    # expand the coords if only one roi is given
    if coords.ndim == 1:
        coords = coords[None, :]

    # manipulate the array
    transform_patch = _apply_struct_mask(
        patch,
        coords=coords,
        struct_params=struct_params,
        rng=rng,
    )
    changed_values = patch[np.where(original_patch != transform_patch)]

    # check that the transformed pixels correspond to the masked pixels
    transformed = []
    axis = -2 + 1 - struct_axis
    for i in range(coords.shape[0]):
        # get indices to mask
        indices_to_mask = [
            c
            for c in range(
                max(0, coords[i, axis] - struct_span // 2),
                min(transform_patch.shape[1] - 1, coords[i, axis] + struct_span // 2)
                + 1,
            )
            if c != coords[i, axis]
        ]

        # add to transform
        if struct_axis == 0:
            transformed.append(
                transform_patch[coords[i, 0], coords[i, 1]][indices_to_mask]
            )
        else:
            transformed.append(
                transform_patch[coords[i, 0], :, coords[i, 2]][indices_to_mask]
            )

    assert np.array_equal(
        np.sort(changed_values), np.sort(np.concatenate(transformed, axis=0))
    )



from pathlib import Path
from threading import Thread
from typing import Tuple

import numpy as np
import pytest
import tifffile
from numpy.typing import NDArray
from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import Callback, EarlyStopping, ModelCheckpoint

from careamics import CAREamist, Configuration, save_configuration
from careamics.config.support import SupportedAlgorithm, SupportedData
from careamics.dataset.dataset_utils import reshape_array
from careamics.lightning.callbacks import HyperParametersCallback, ProgressBarCallback
from careamics.lightning.predict_data_module import create_predict_datamodule


def random_array(shape: Tuple[int, ...], seed: int = 42):
    """Return a random array with values between 0 and 255."""
    rng = np.random.default_rng(seed)
    return (rng.integers(0, 255, shape)).astype(np.float32)


def test_no_parameters():
    """Test that CAREamics cannot be instantiated without parameters."""
    with pytest.raises(TypeError):
        CAREamist()


def test_minimum_configuration_via_object(tmp_path: Path, minimum_configuration: dict):
    """Test that CAREamics can be instantiated with a minimum configuration object."""
    # create configuration
    config = Configuration(**minimum_configuration)

    # instantiate CAREamist
    CAREamist(source=config, work_dir=tmp_path)


def test_minimum_configuration_via_path(tmp_path: Path, minimum_configuration: dict):
    """Test that CAREamics can be instantiated with a path to a minimum
    configuration.
    """
    # create configuration
    config = Configuration(**minimum_configuration)
    path_to_config = save_configuration(config, tmp_path)

    # instantiate CAREamist
    CAREamist(source=path_to_config)


def test_train_error_target_unsupervised_algorithm(
    tmp_path: Path, minimum_configuration: dict
):
    """Test that an error is raised when a target is provided for N2V."""
    # create configuration
    config = Configuration(**minimum_configuration)
    config.algorithm_config.algorithm = SupportedAlgorithm.N2V.value

    # train error with Paths
    config.data_config.data_type = SupportedData.TIFF.value
    careamics = CAREamist(source=config, work_dir=tmp_path)
    with pytest.raises(ValueError):
        careamics.train(
            train_source=tmp_path,
            train_target=tmp_path,
        )

    # train error with strings
    with pytest.raises(ValueError):
        careamics.train(
            train_source=str(tmp_path),
            train_target=str(tmp_path),
        )

    # train error with arrays
    config.data_config.data_type = SupportedData.ARRAY.value
    careamics = CAREamist(source=config, work_dir=tmp_path)
    with pytest.raises(ValueError):
        careamics.train(
            train_source=np.ones((32, 32)),
            train_target=np.ones((32, 32)),
        )


@pytest.mark.skip(reason="bmz")
def test_train_single_array_no_val(tmp_path: Path, minimum_configuration: dict):
    """Test that CAREamics can be trained with arrays."""
    # training data
    train_array = random_array((32, 32))

    # create configuration
    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "YX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.ARRAY.value
    config.data_config.patch_size = (8, 8)

    # instantiate CAREamist
    careamist = CAREamist(source=config, work_dir=tmp_path)

    # train CAREamist
    careamist.train(train_source=train_array)

    # check that it trained
    assert Path(tmp_path / "checkpoints" / "last.ckpt").exists()

    # export to BMZ
    careamist.export_to_bmz(
        path_to_archive=tmp_path / "model.zip",
        friendly_model_name="TopModel",
        input_array=train_array,
        authors=[{"name": "Amod", "affiliation": "El"}],
        general_description="A model that just walked in.",
        data_description="A random array.",
    )
    assert (tmp_path / "model.zip").exists()


@pytest.mark.skip(reason="bmz")
def test_train_array(tmp_path: Path, minimum_configuration: dict):
    """Test that CAREamics can be trained on arrays."""
    # training data
    train_array = random_array((32, 32))
    val_array = random_array((32, 32))

    # create configuration
    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "YX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.ARRAY.value
    config.data_config.patch_size = (8, 8)

    # instantiate CAREamist
    careamist = CAREamist(source=config, work_dir=tmp_path)

    # train CAREamist
    careamist.train(train_source=train_array, val_source=val_array)

    # check that it trained
    assert Path(tmp_path / "checkpoints" / "last.ckpt").exists()

    # export to BMZ
    careamist.export_to_bmz(
        path_to_archive=tmp_path / "model.zip",
        friendly_model_name="TopModel",
        input_array=train_array,
        authors=[{"name": "Amod", "affiliation": "El"}],
        general_description="A model that just walked in.",
        data_description="A random array.",
    )
    assert (tmp_path / "model.zip").exists()


@pytest.mark.skip(reason="bmz")
@pytest.mark.parametrize("independent_channels", [False, True])
def test_train_array_channel(
    tmp_path: Path, minimum_configuration: dict, independent_channels: bool
):
    """Test that CAREamics can be trained on arrays with channels."""
    # training data
    train_array = random_array((32, 32, 3))
    val_array = random_array((32, 32, 3))

    # create configuration
    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "YXC"
    config.algorithm_config.model.in_channels = 3
    config.algorithm_config.model.num_classes = 3
    config.algorithm_config.model.independent_channels = independent_channels
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.ARRAY.value
    config.data_config.patch_size = (8, 8)

    # instantiate CAREamist
    careamist = CAREamist(source=config, work_dir=tmp_path)

    # train CAREamist
    careamist.train(train_source=train_array, val_source=val_array)

    # check that it trained
    assert Path(tmp_path / "checkpoints" / "last.ckpt").exists()

    # export to BMZ
    careamist.export_to_bmz(
        path_to_archive=tmp_path / "model.zip",
        friendly_model_name="TopModel",
        input_array=train_array,
        authors=[{"name": "Amod", "affiliation": "El"}],
        general_description="A model that just walked in.",
        data_description="A random array.",
        channel_names=["red", "green", "blue"],
    )
    assert (tmp_path / "model.zip").exists()


@pytest.mark.skip(reason="bmz")
def test_train_array_3d(tmp_path: Path, minimum_configuration: dict):
    """Test that CAREamics can be trained on 3D arrays."""
    # training data
    train_array = random_array((8, 32, 32))
    val_array = random_array((8, 32, 32))

    # create configuration
    minimum_configuration["data_config"]["axes"] = "ZYX"
    minimum_configuration["data_config"]["patch_size"] = (8, 16, 16)
    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 1
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.ARRAY.value

    # instantiate CAREamist
    careamist = CAREamist(source=config, work_dir=tmp_path)

    # train CAREamist
    careamist.train(train_source=train_array, val_source=val_array)

    # check that it trained
    assert Path(tmp_path / "checkpoints" / "last.ckpt").exists()

    # export to BMZ
    careamist.export_to_bmz(
        path_to_archive=tmp_path / "model.zip",
        friendly_model_name="TopModel",
        input_array=train_array,
        authors=[{"name": "Amod", "affiliation": "El"}],
        general_description="A model that just walked in.",
        data_description="A random array.",
    )
    assert (tmp_path / "model.zip").exists()


@pytest.mark.skip(reason="bmz")
def test_train_tiff_files_in_memory_no_val(tmp_path: Path, minimum_configuration: dict):
    """Test that CAREamics can be trained with tiff files in memory."""
    # training data
    train_array = random_array((32, 32))

    # save files
    train_file = tmp_path / "train.tiff"
    tifffile.imwrite(train_file, train_array)

    # create configuration
    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "YX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.TIFF.value
    config.data_config.patch_size = (8, 8)

    # instantiate CAREamist
    careamist = CAREamist(source=config, work_dir=tmp_path)

    # train CAREamist
    careamist.train(train_source=train_file)

    # check that it trained
    assert Path(tmp_path / "checkpoints" / "last.ckpt").exists()

    # export to BMZ
    careamist.export_to_bmz(
        path_to_archive=tmp_path / "model.zip",
        friendly_model_name="TopModel",
        input_array=train_array,
        authors=[{"name": "Amod", "affiliation": "El"}],
        general_description="A model that just walked in.",
        data_description="A random array.",
    )
    assert (tmp_path / "model.zip").exists()


@pytest.mark.skip(reason="bmz")
def test_train_tiff_files_in_memory(tmp_path: Path, minimum_configuration: dict):
    """Test that CAREamics can be trained with tiff files in memory."""
    # training data
    train_array = random_array((32, 32))
    val_array = random_array((32, 32))

    # save files
    train_file = tmp_path / "train.tiff"
    tifffile.imwrite(train_file, train_array)

    val_file = tmp_path / "val.tiff"
    tifffile.imwrite(val_file, val_array)

    # create configuration
    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "YX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.TIFF.value
    config.data_config.patch_size = (8, 8)

    # instantiate CAREamist
    careamist = CAREamist(source=config, work_dir=tmp_path)

    # train CAREamist
    careamist.train(train_source=train_file, val_source=val_file)

    # check that it trained
    assert Path(tmp_path / "checkpoints" / "last.ckpt").exists()

    # export to BMZ
    careamist.export_to_bmz(
        path_to_archive=tmp_path / "model.zip",
        friendly_model_name="TopModel",
        input_array=train_array,
        authors=[{"name": "Amod", "affiliation": "El"}],
        general_description="A model that just walked in.",
        data_description="A random array.",
    )
    assert (tmp_path / "model.zip").exists()


def test_train_tiff_files(tmp_path: Path, minimum_configuration: dict):
    """Test that CAREamics can be trained with tiff files by deactivating
    the in memory dataset.
    """
    # training data
    train_array = random_array((32, 32))
    val_array = random_array((32, 32))

    # save files
    train_file = tmp_path / "train.tiff"
    tifffile.imwrite(train_file, train_array)

    val_file = tmp_path / "val.tiff"
    tifffile.imwrite(val_file, val_array)

    # create configuration
    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "YX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.TIFF.value
    config.data_config.patch_size = (8, 8)

    # instantiate CAREamist
    careamist = CAREamist(source=config, work_dir=tmp_path)

    # train CAREamist
    careamist.train(train_source=train_file, val_source=val_file, use_in_memory=False)

    # check that it trained
    assert Path(tmp_path / "checkpoints" / "last.ckpt").exists()

    # export to BMZ
    careamist.export_to_bmz(
        path_to_archive=tmp_path / "model.zip",
        friendly_model_name="TopModel",
        input_array=train_array,
        authors=[{"name": "Amod", "affiliation": "El"}],
        general_description="A model that just walked in.",
        data_description="A random array.",
    )
    assert (tmp_path / "model.zip").exists()


@pytest.mark.skip(reason="bmz")
def test_train_array_supervised(tmp_path: Path, supervised_configuration: dict):
    """Test that CAREamics can be trained with arrays."""
    # training data
    train_array = random_array((32, 32))
    val_array = random_array((32, 32))
    train_target = random_array((32, 32))
    val_target = random_array((32, 32))

    # create configuration
    config = Configuration(**supervised_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "YX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.ARRAY.value
    config.data_config.patch_size = (8, 8)

    # instantiate CAREamist
    careamist = CAREamist(source=config, work_dir=tmp_path)

    # train CAREamist
    careamist.train(
        train_source=train_array,
        val_source=val_array,
        train_target=train_target,
        val_target=val_target,
    )

    # check that it trained
    assert Path(tmp_path / "checkpoints" / "last.ckpt").exists()

    # export to BMZ
    careamist.export_to_bmz(
        path_to_archive=tmp_path / "model.zip",
        friendly_model_name="TopModel",
        input_array=train_array,
        authors=[{"name": "Amod", "affiliation": "El"}],
        general_description="A model that just walked in.",
        data_description="A random array.",
    )
    assert (tmp_path / "model.zip").exists()


@pytest.mark.skip(reason="bmz")
def test_train_tiff_files_in_memory_supervised(
    tmp_path: Path, supervised_configuration: dict
):
    """Test that CAREamics can be trained with tiff files in memory."""
    # training data
    train_array = random_array((32, 32))
    val_array = random_array((32, 32))
    train_target = random_array((32, 32))
    val_target = random_array((32, 32))

    # save files
    images = tmp_path / "images"
    images.mkdir()
    train_file = images / "train.tiff"
    tifffile.imwrite(train_file, train_array)

    val_file = tmp_path / "images" / "val.tiff"
    tifffile.imwrite(val_file, val_array)

    targets = tmp_path / "targets"
    targets.mkdir()
    train_target_file = targets / "train.tiff"
    tifffile.imwrite(train_target_file, train_target)

    val_target_file = targets / "val.tiff"
    tifffile.imwrite(val_target_file, val_target)

    # create configuration
    config = Configuration(**supervised_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "YX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.TIFF.value
    config.data_config.patch_size = (8, 8)

    # instantiate CAREamist
    careamist = CAREamist(source=config, work_dir=tmp_path)

    # train CAREamist
    careamist.train(
        train_source=train_file,
        val_source=val_file,
        train_target=train_target_file,
        val_target=val_target_file,
    )

    # check that it trained
    assert Path(tmp_path / "checkpoints" / "last.ckpt").exists()

    # export to BMZ
    careamist.export_to_bmz(
        path_to_archive=tmp_path / "model.zip",
        friendly_model_name="TopModel",
        input_array=train_array,
        authors=[{"name": "Amod", "affiliation": "El"}],
        general_description="A model that just walked in.",
        data_description="A random array.",
    )
    assert (tmp_path / "model.zip").exists()


def test_train_tiff_files_supervised(tmp_path: Path, supervised_configuration: dict):
    """Test that CAREamics can be trained with tiff files by deactivating
    the in memory dataset.
    """
    # training data
    train_array = random_array((32, 32))
    val_array = random_array((32, 32))
    train_target = random_array((32, 32))
    val_target = random_array((32, 32))

    # save files
    images = tmp_path / "images"
    images.mkdir()
    train_file = images / "train.tiff"
    tifffile.imwrite(train_file, train_array)

    val_file = tmp_path / "images" / "val.tiff"
    tifffile.imwrite(val_file, val_array)

    targets = tmp_path / "targets"
    targets.mkdir()
    train_target_file = targets / "train.tiff"
    tifffile.imwrite(train_target_file, train_target)

    val_target_file = targets / "val.tiff"
    tifffile.imwrite(val_target_file, val_target)

    # create configuration
    config = Configuration(**supervised_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "YX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.TIFF.value
    config.data_config.patch_size = (8, 8)

    # instantiate CAREamist
    careamist = CAREamist(source=config, work_dir=tmp_path)

    # train CAREamist
    careamist.train(
        train_source=train_file,
        val_source=val_file,
        train_target=train_target_file,
        val_target=val_target_file,
        use_in_memory=False,
    )

    # check that it trained
    assert Path(tmp_path / "checkpoints" / "last.ckpt").exists()

    # export to BMZ
    careamist.export_to_bmz(
        path_to_archive=tmp_path / "model.zip",
        friendly_model_name="TopModel",
        input_array=train_array,
        authors=[{"name": "Amod", "affiliation": "El"}],
        general_description="A model that just walked in.",
        data_description="A random array.",
    )
    assert (tmp_path / "model.zip").exists()


@pytest.mark.skip(reason="bmz")
@pytest.mark.parametrize("samples", [1, 2, 4])
@pytest.mark.parametrize("batch_size", [1, 2])
def test_predict_on_array_tiled(
    tmp_path: Path, minimum_configuration: dict, batch_size, samples
):
    """Test that CAREamics can predict on arrays."""
    # training data
    train_array = random_array((samples, 32, 32))

    # create configuration
    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "SYX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.ARRAY.value
    config.data_config.patch_size = (8, 8)

    # instantiate CAREamist
    careamist = CAREamist(source=config, work_dir=tmp_path)

    # train CAREamist
    careamist.train(train_source=train_array)

    # predict CAREamist
    predicted = careamist.predict(
        train_array, batch_size=batch_size, tile_size=(16, 16), tile_overlap=(4, 4)
    )

    assert (
        np.concatenate(predicted).shape
        == reshape_array(train_array, config.data_config.axes).shape
    )

    # export to BMZ
    careamist.export_to_bmz(
        path_to_archive=tmp_path / "model.zip",
        friendly_model_name="TopModel",
        input_array=train_array,
        authors=[{"name": "Amod", "affiliation": "El"}],
        general_description="A model that just walked in.",
        data_description="A random array.",
    )
    assert (tmp_path / "model.zip").exists()


@pytest.mark.skip(reason="bmz")
@pytest.mark.parametrize("samples", [1, 2, 4])
@pytest.mark.parametrize("batch_size", [1, 2])
def test_predict_arrays_no_tiling(
    tmp_path: Path, minimum_configuration: dict, batch_size, samples
):
    """Test that CAREamics can predict on arrays without tiling."""
    # training data
    train_array = random_array((samples, 32, 32))

    # create configuration
    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "SYX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.ARRAY.value
    config.data_config.patch_size = (8, 8)

    # instantiate CAREamist
    careamist = CAREamist(source=config, work_dir=tmp_path)

    # train CAREamist
    careamist.train(train_source=train_array)

    # predict CAREamist
    predicted = careamist.predict(train_array, batch_size=batch_size)

    assert (
        np.concatenate(predicted).shape
        == reshape_array(train_array, config.data_config.axes).shape
    )

    # export to BMZ
    careamist.export_to_bmz(
        path_to_archive=tmp_path / "model.zip",
        friendly_model_name="TopModel",
        input_array=train_array,
        authors=[{"name": "Amod", "affiliation": "El"}],
        general_description="A model that just walked in.",
        data_description="A random array.",
    )
    assert (tmp_path / "model.zip").exists()


@pytest.mark.skip(
    reason=(
        "This might be a problem at the PyTorch level during `forward`. Values up to "
        "0.001 different."
    )
)
def test_batched_prediction(tmp_path: Path, minimum_configuration: dict):
    "Compare outputs when a batch size of 1 or 2 is used"

    tile_size = (16, 16)
    tile_overlap = (4, 4)
    shape = (32, 32)

    train_array = random_array(shape)
    # create configuration
    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "YX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.ARRAY.value

    # instantiate CAREamist
    careamist = CAREamist(source=config, work_dir=tmp_path)

    # train CAREamist
    careamist.train(train_source=train_array)

    # predict with batch size 1 and batch size 2
    pred_bs_1 = careamist.predict(
        train_array, batch_size=1, tile_size=tile_size, tile_overlap=tile_overlap
    )
    pred_bs_2 = careamist.predict(
        train_array, batch_size=2, tile_size=tile_size, tile_overlap=tile_overlap
    )

    assert np.array_equal(pred_bs_1, pred_bs_2)


@pytest.mark.parametrize("independent_channels", [False, True])
@pytest.mark.parametrize("batch_size", [1, 2])
def test_predict_tiled_channel(
    tmp_path: Path,
    minimum_configuration: dict,
    independent_channels: bool,
    batch_size: int,
):
    """Test that CAREamics can be trained on arrays with channels."""
    # training data
    train_array = random_array((3, 32, 32))
    val_array = random_array((3, 32, 32))

    # create configuration
    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "CYX"
    config.algorithm_config.model.in_channels = 3
    config.algorithm_config.model.num_classes = 3
    config.algorithm_config.model.independent_channels = independent_channels
    config.data_config.batch_size = batch_size
    config.data_config.data_type = SupportedData.ARRAY.value
    config.data_config.patch_size = (8, 8)

    # instantiate CAREamist
    careamist = CAREamist(source=config, work_dir=tmp_path)

    # train CAREamist
    careamist.train(train_source=train_array, val_source=val_array)

    # predict CAREamist
    predicted = careamist.predict(
        train_array, batch_size=batch_size, tile_size=(16, 16), tile_overlap=(4, 4)
    )

    assert (
        np.concatenate(predicted).shape
        == reshape_array(train_array, config.data_config.axes).shape
    )


@pytest.mark.skip(reason="bmz")
@pytest.mark.parametrize("tiled", [True, False])
@pytest.mark.parametrize("n_samples", [1, 2])
@pytest.mark.parametrize("batch_size", [1, 2])
def test_predict_path(
    tmp_path: Path, minimum_configuration: dict, batch_size, n_samples, tiled
):
    """Test that CAREamics can predict with tiff files."""
    # training data
    train_array = random_array((32, 32))

    # save files
    for i in range(n_samples):
        train_file = tmp_path / f"train_{i}.tiff"
        tifffile.imwrite(train_file, train_array)

    # create configuration
    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "YX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.TIFF.value
    config.data_config.patch_size = (8, 8)

    # instantiate CAREamist
    careamist = CAREamist(source=config, work_dir=tmp_path)

    # train CAREamist
    careamist.train(train_source=train_file)

    if tiled:
        tile_size = (16, 16)
        tile_overlap = (4, 4)
    else:
        tile_size = None
        tile_overlap = None

    # predict CAREamist
    predicted = careamist.predict(
        train_file,
        batch_size=batch_size,
        tile_size=tile_size,
        tile_overlap=tile_overlap,
    )

    # check that it predicted
    if isinstance(predicted, list):
        for p in predicted:
            assert p.squeeze().shape == train_array.shape
    else:
        assert predicted.squeeze().shape == train_array.shape

    # export to BMZ
    careamist.export_to_bmz(
        path_to_archive=tmp_path / "model.zip",
        friendly_model_name="TopModel",
        input_array=train_array,
        authors=[{"name": "Amod", "affiliation": "El"}],
        general_description="A model that just walked in.",
        data_description="A random array.",
    )
    assert (tmp_path / "model.zip").exists()


def test_predict_pretrained_checkpoint(tmp_path: Path, pre_trained: Path):
    """Test that CAREamics can be instantiated with a pre-trained network and predict
    on an array."""
    # prediction data
    source_array = random_array((32, 32))

    # instantiate CAREamist
    careamist = CAREamist(source=pre_trained, work_dir=tmp_path)
    assert careamist.cfg.data_config.image_means is not None
    assert careamist.cfg.data_config.image_stds is not None

    # predict
    predicted = careamist.predict(source_array)

    # check that it predicted
    assert (
        np.concatenate(predicted).shape
        == reshape_array(source_array, careamist.cfg.data_config.axes).shape
    )


def test_predict_pretrained_bmz(tmp_path: Path, pre_trained_bmz: Path):
    """Test that CAREamics can be instantiated with a BMZ archive and predict."""
    # prediction data
    source_array = random_array((32, 32))

    # instantiate CAREamist
    careamist = CAREamist(source=pre_trained_bmz, work_dir=tmp_path)

    # predict
    predicted = careamist.predict(source_array)

    # check that it predicted
    assert (
        np.concatenate(predicted).shape
        == reshape_array(source_array, careamist.cfg.data_config.axes).shape
    )


def test_export_bmz_pretrained_prediction(tmp_path: Path, pre_trained: Path):
    """Test that CAREamics can be instantiated with a pre-trained network and exported
    to BMZ after prediction.

    In this case, the careamist extracts the BMZ test data from the prediction
    datamodule.
    """
    # instantiate CAREamist
    careamist = CAREamist(source=pre_trained, work_dir=tmp_path)

    # prediction data
    source_array = random_array((32, 32))
    _ = careamist.predict(source_array)
    assert len(careamist.pred_datamodule.predict_dataloader()) > 0

    # export to BMZ (random array created)
    careamist.export_to_bmz(
        path_to_archive=tmp_path / "model.zip",
        friendly_model_name="TopModel",
        input_array=source_array,
        authors=[{"name": "Amod", "affiliation": "El"}],
        general_description="A model that just walked in.",
        data_description="A random array.",
    )
    assert (tmp_path / "model.zip").exists()


def test_export_bmz_pretrained_with_array(tmp_path: Path, pre_trained: Path):
    """Test that CAREamics can be instantiated with a pre-trained network and exported
    to BMZ.

    In this case, we provide an array to the BMZ archive test.
    """
    # instantiate CAREamist
    careamist = CAREamist(source=pre_trained, work_dir=tmp_path)

    # alternatively we can pass an array
    array = random_array((32, 32))
    careamist.export_to_bmz(
        path_to_archive=tmp_path / "model2.zip",
        friendly_model_name="TopModel",
        input_array=array,
        authors=[{"name": "Amod", "affiliation": "El"}],
        general_description="A model that just walked in.",
        data_description="A random array.",
    )
    assert (tmp_path / "model2.zip").exists()


def test_predict_to_disk_path_tiff(tmp_path, minimum_configuration):
    """Test predict_to_disk function with path source and tiff write type."""

    # prepare dummy data
    train_array = random_array((32, 32))

    image_dir = tmp_path / "images"
    image_dir.mkdir()
    n_samples = 2
    # save files
    for i in range(n_samples):
        train_file = image_dir / f"image_{i}.tiff"
        tifffile.imwrite(train_file, train_array)

    # create configuration
    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "YX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.TIFF.value
    config.data_config.patch_size = (8, 8)

    # train
    careamist = CAREamist(source=config, work_dir=tmp_path)
    careamist.train(train_source=image_dir)

    # predict to disk
    careamist.predict_to_disk(source=image_dir)

    for i in range(n_samples):
        assert (tmp_path / "predictions" / f"image_{i}.tiff").is_file()


def test_predict_to_disk_datamodule_tiff(tmp_path, minimum_configuration):
    """Test predict_to_disk function with datamodule source and tiff write type."""

    # prepare dummy data
    train_array = random_array((32, 32))

    image_dir = tmp_path / "images"
    image_dir.mkdir()
    n_samples = 2
    # save files
    for i in range(n_samples):
        train_file = image_dir / f"image_{i}.tiff"
        tifffile.imwrite(train_file, train_array)

    # create configuration
    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "YX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.TIFF.value
    config.data_config.patch_size = (8, 8)

    # train
    careamist = CAREamist(source=config, work_dir=tmp_path)
    careamist.train(train_source=image_dir)

    datamodule = create_predict_datamodule(
        pred_data=image_dir,
        data_type=config.data_config.data_type,
        axes=config.data_config.axes,
        image_means=careamist.cfg.data_config.image_means,
        image_stds=careamist.cfg.data_config.image_stds,
    )

    # predict to disk
    careamist.predict_to_disk(source=datamodule)

    for i in range(n_samples):
        assert (tmp_path / "predictions" / f"image_{i}.tiff").is_file()


def test_predict_to_disk_custom(tmp_path, minimum_configuration):
    """Test predict_to_disk function with custom write type."""

    def write_numpy(file_path: Path, img: NDArray, *args, **kwargs) -> None:
        np.save(file=file_path, arr=img)

    # prepare dummy data
    train_array = random_array((32, 32))

    image_dir = tmp_path / "images"
    image_dir.mkdir()
    n_samples = 2
    # save files
    for i in range(n_samples):
        train_file = image_dir / f"image_{i}.tiff"
        tifffile.imwrite(train_file, train_array)

    # create configuration
    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "YX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.TIFF.value
    config.data_config.patch_size = (8, 8)

    # train
    careamist = CAREamist(source=config, work_dir=tmp_path)
    careamist.train(train_source=image_dir)

    # predict to disk
    careamist.predict_to_disk(
        source=image_dir,
        write_type=SupportedData.CUSTOM,
        write_extension=".npy",
        write_func=write_numpy,
    )

    for i in range(n_samples):
        assert (tmp_path / "predictions" / f"image_{i}.npy").is_file()


def test_predict_to_disk_custom_raises(tmp_path, minimum_configuration):
    """
    Test predict_to_disk custom write type raises ValueError.

    ValueError should be raised if no write_extension or no write_func is provided.
    """

    def write_numpy(file_path: Path, img: NDArray, *args, **kwargs) -> None:
        np.save(file=file_path, arr=img)

    # prepare dummy data
    train_array = random_array((32, 32))

    image_dir = tmp_path / "images"
    image_dir.mkdir()
    n_samples = 2
    # save files
    for i in range(n_samples):
        train_file = image_dir / f"image_{i}.tiff"
        tifffile.imwrite(train_file, train_array)

    # create configuration
    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "YX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.TIFF.value
    config.data_config.patch_size = (8, 8)

    # train
    careamist = CAREamist(source=config, work_dir=tmp_path)
    careamist.train(train_source=image_dir)

    with pytest.raises(ValueError):
        # no write extension provided
        careamist.predict_to_disk(
            source=image_dir,
            write_type=SupportedData.CUSTOM,
            write_extension=None,
            write_func=write_numpy,
        )
    with pytest.raises(ValueError):
        # no write func provided.
        careamist.predict_to_disk(
            source=image_dir,
            write_type=SupportedData.CUSTOM,
            write_extension=".npy",
            write_func=None,
        )


def test_add_custom_callback(tmp_path, minimum_configuration):
    """Test that custom callback can be added to the CAREamist."""

    # define a custom callback
    class MyCallback(Callback):
        def __init__(self):
            super().__init__()

            self.has_started = False
            self.has_ended = False

        def on_train_start(self, trainer, pl_module):
            self.has_started = True

        def on_train_end(self, trainer, pl_module):
            self.has_ended = True

    my_callback = MyCallback()
    assert not my_callback.has_started
    assert not my_callback.has_ended

    # training data
    train_array = random_array((32, 32))

    # create configuration
    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "YX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.ARRAY.value
    config.data_config.patch_size = (8, 8)

    # instantiate CAREamist
    careamist = CAREamist(source=config, work_dir=tmp_path, callbacks=[my_callback])
    assert not my_callback.has_started
    assert not my_callback.has_ended

    # train CAREamist
    careamist.train(train_source=train_array)

    # check the state of the callback
    assert my_callback.has_started
    assert my_callback.has_ended


def test_error_passing_careamics_callback(tmp_path, minimum_configuration):
    """Test that an error is thrown if we pass known callbacks to CAREamist."""
    # create configuration
    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 1
    config.data_config.axes = "YX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.ARRAY.value
    config.data_config.patch_size = (8, 8)

    # Lightning callbacks
    model_ckp = ModelCheckpoint()

    with pytest.raises(ValueError):
        CAREamist(source=config, work_dir=tmp_path, callbacks=[model_ckp])

    early_stp = EarlyStopping(
        Trainer(
            max_epochs=1,
            default_root_dir=tmp_path,
        )
    )

    with pytest.raises(ValueError):
        CAREamist(source=config, work_dir=tmp_path, callbacks=[early_stp])

    # CAREamics callbacks
    progress_bar = ProgressBarCallback()

    with pytest.raises(ValueError):
        CAREamist(source=config, work_dir=tmp_path, callbacks=[progress_bar])

    hyper_params = HyperParametersCallback(config=config)

    with pytest.raises(ValueError):
        CAREamist(source=config, work_dir=tmp_path, callbacks=[hyper_params])


def test_stop_training(tmp_path: Path, minimum_configuration: dict):
    """Test that CAREamics can stop the training"""
    # training data
    train_array = random_array((32, 32))

    # create configuration
    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 1_000
    config.data_config.axes = "YX"
    config.data_config.batch_size = 2
    config.data_config.data_type = SupportedData.ARRAY.value
    config.data_config.patch_size = (8, 8)

    # instantiate CAREamist
    careamist = CAREamist(source=config, work_dir=tmp_path)

    def _train():
        careamist.train(train_source=train_array)

    # create thread
    thread = Thread(target=_train)
    thread.start()

    # stop training
    careamist.stop_training()
    thread.join()

    assert careamist.trainer.should_stop


def test_read_logger(tmp_path, minimum_configuration):

    config = Configuration(**minimum_configuration)
    config.training_config.num_epochs = 10

    array = np.arange(32 * 32).reshape((32, 32))

    careamist = CAREamist(config, work_dir=tmp_path)
    careamist.train(train_source=array)
    losses = careamist.get_losses()

    assert len(losses) == 4
    for key in losses:
        assert len(losses[key]) == config.training_config.num_epochs



import numpy as np
import pytest

from careamics.dataset.patching.random_patching import extract_patches_random


@pytest.mark.parametrize(
    "shape, patch_size",
    [
        ((1, 1, 8, 8), (3, 3)),
        ((1, 3, 8, 8), (3, 3)),
        ((3, 1, 8, 8), (3, 3)),
        ((2, 3, 8, 8), (3, 3)),
        ((1, 1, 5, 8, 8), (3, 3, 3)),
        ((1, 3, 5, 8, 8), (3, 3, 3)),
        ((3, 1, 5, 8, 8), (3, 3, 3)),
        ((2, 3, 5, 8, 8), (3, 3, 3)),
    ],
)
def test_random_patching_unsupervised(ordered_array, shape, patch_size):
    """Check that the patches are extracted correctly.

    Since extract patches is called on already shaped array, dimensions S and C are
    present.
    """
    # create array
    array = ordered_array(shape)
    is_3D = len(patch_size) == 3

    # minimum number of unique patches to validate randomness
    if is_3D:
        min_unique_patches = 0.7 * np.prod(shape[-3:]) / np.prod(patch_size)
    else:
        min_unique_patches = 0.7 * np.prod(shape[-2:]) / np.prod(patch_size)

    top_left = []
    seeds = [24, 42, 56]
    for i in range(3):
        patch_generator = extract_patches_random(
            array, patch_size=patch_size, seed=seeds[i]
        )

        # get all patches and targets
        patches = [patch for patch, _ in patch_generator]

        # check patch shape
        for patch in patches:
            # account for C dimension
            assert patch.shape[1:] == patch_size

            # get top_left index in the original array
            if is_3D:
                ind = np.where(array == patch[0, 0, 0, 0])
            else:
                ind = np.where(array == patch[0, 0, 0])

            top_left.append(np.array(ind))

        # check randomness
        coords = np.array(top_left).squeeze()
        assert coords.min() == 0
        assert coords.max() == max(array.shape) - max(patch_size)
        assert len(np.unique(coords, axis=0)) >= min_unique_patches



from pathlib import Path

import numpy as np
from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import (
    ModelCheckpoint,
)

from careamics import Configuration
from careamics.lightning import (
    create_careamics_module,
    create_predict_datamodule,
    create_train_datamodule,
)
from careamics.prediction_utils import convert_outputs


def test_smoke_n2v_2d_array(tmp_path, minimum_configuration):
    """Test a full run of N2V training with the lightning API."""
    rng = np.random.default_rng(42)

    # training data
    train_array = rng.integers(0, 255, (32, 32)).astype(np.float32)
    val_array = rng.integers(0, 255, (32, 32)).astype(np.float32)

    cfg = Configuration(**minimum_configuration)

    # create lightning module
    model = create_careamics_module(
        algorithm=cfg.algorithm_config.algorithm,
        loss=cfg.algorithm_config.loss,
        architecture=cfg.algorithm_config.model.architecture,
    )

    # create data module
    data = create_train_datamodule(
        train_data=train_array,
        val_data=val_array,
        data_type=cfg.data_config.data_type,
        patch_size=cfg.data_config.patch_size,
        axes=cfg.data_config.axes,
        batch_size=cfg.data_config.batch_size,
    )

    # create trainer
    trainer = Trainer(
        max_epochs=1,
        default_root_dir=tmp_path,
        callbacks=[
            ModelCheckpoint(
                dirpath=tmp_path / "checkpoints",
                filename="test_lightning_api",
            )
        ],
    )

    # train
    trainer.fit(model, datamodule=data)
    assert Path(tmp_path / "checkpoints" / "test_lightning_api.ckpt").exists()

    # predict
    means, stds = data.get_data_statistics()
    predict_data = create_predict_datamodule(
        pred_data=val_array,
        data_type=cfg.data_config.data_type,
        axes=cfg.data_config.axes,
        image_means=means,
        image_stds=stds,
    )

    # predict
    predicted = trainer.predict(model, datamodule=predict_data)
    assert predicted[0].squeeze().shape == val_array.shape


def test_smoke_n2v_2d_tiling(tmp_path, minimum_configuration):
    """Test a full run of N2V training with the lightning API and tiled prediction."""
    # training data
    rng = np.random.default_rng(42)
    train_array = rng.integers(0, 255, (32, 32)).astype(np.float32)
    val_array = rng.integers(0, 255, (32, 32)).astype(np.float32)

    cfg = Configuration(**minimum_configuration)

    # create lightning module
    model = create_careamics_module(
        algorithm=cfg.algorithm_config.algorithm,
        loss=cfg.algorithm_config.loss,
        architecture=cfg.algorithm_config.model.architecture,
    )

    # create data module
    data = create_train_datamodule(
        train_data=train_array,
        val_data=val_array,
        data_type=cfg.data_config.data_type,
        patch_size=cfg.data_config.patch_size,
        axes=cfg.data_config.axes,
        batch_size=cfg.data_config.batch_size,
    )

    # create trainer
    trainer = Trainer(
        max_epochs=1,
        default_root_dir=tmp_path,
        callbacks=[
            ModelCheckpoint(
                dirpath=tmp_path / "checkpoints",
                filename="test_lightning_api",
            )
        ],
    )

    # train
    trainer.fit(model, datamodule=data)
    assert Path(tmp_path / "checkpoints" / "test_lightning_api.ckpt").exists()

    # predict
    means, stds = data.get_data_statistics()
    predict_data = create_predict_datamodule(
        pred_data=val_array,
        data_type=cfg.data_config.data_type,
        axes=cfg.data_config.axes,
        image_means=means,
        image_stds=stds,
        tile_size=(8, 8),
        tile_overlap=(2, 2),
    )

    # predict
    predicted = trainer.predict(model, datamodule=predict_data)
    predicted_stitched = convert_outputs(predicted, tiled=True)
    assert predicted_stitched[0].shape[-2:] == val_array.shape



import numpy as np


class IndexSwitcher:
    """
    The idea is to switch from valid indices for target to invalid indices for target.
    If index in invalid for the target, then we return all zero vector as target.
    This combines both logic:
    1. Using less amount of total data.
    2. Using less amount of target data but using full data.
    """

    def __init__(self, idx_manager, data_config, patch_size) -> None:
        self.idx_manager = idx_manager
        self._data_shape = self.idx_manager.get_data_shape()
        self._training_validtarget_fraction = data_config.get(
            "training_validtarget_fraction", 1.0
        )
        self._validtarget_ceilT = int(
            np.ceil(self._data_shape[0] * self._training_validtarget_fraction)
        )
        self._patch_size = patch_size
        assert (
            data_config.deterministic_grid is True
        ), "This only works when the dataset has deterministic grid. Needed randomness comes from this class."
        assert (
            "grid_size" in data_config and data_config.grid_size == 1
        ), "We need a one to one mapping between index and h, w, t"

        self._h_validmax, self._w_validmax = self.get_reduced_frame_size(
            self._data_shape[:3], self._training_validtarget_fraction
        )
        if self._h_validmax < self._patch_size or self._w_validmax < self._patch_size:
            print(
                "WARNING: The valid target size is smaller than the patch size. This will result in all zero target. so, we are ignoring this frame for target."
            )
            self._h_validmax = 0
            self._w_validmax = 0

        print(
            f"[{self.__class__.__name__}] Target Indices: [0,{self._validtarget_ceilT - 1}]. Index={self._validtarget_ceilT - 1} has shape [:{self._h_validmax},:{self._w_validmax}].  Available data: {self._data_shape[0]}"
        )

    def get_valid_target_index(self):
        """
        Returns an index which corresponds to a frame which is expected to have a target.
        """
        _, h, w, _ = self._data_shape
        framepixelcount = h * w
        targetpixels = np.array(
            [framepixelcount] * (self._validtarget_ceilT - 1)
            + [self._h_validmax * self._w_validmax]
        )
        targetpixels = targetpixels / np.sum(targetpixels)
        t = np.random.choice(self._validtarget_ceilT, p=targetpixels)
        # t = np.random.randint(0, self._validtarget_ceilT) if self._validtarget_ceilT >= 1 else 0
        h, w = self.get_valid_target_hw(t)
        index = self.idx_manager.idx_from_hwt(h, w, t)
        # print('Valid', index, h,w,t)
        return index

    def get_invalid_target_index(self):
        # if self._validtarget_ceilT == 0:
        # TODO: There may not be enough data for this to work. The better way is to skip using 0 for invalid target.
        # t = np.random.randint(1, self._data_shape[0])
        # elif self._validtarget_ceilT < self._data_shape[0]:
        #     t = np.random.randint(self._validtarget_ceilT, self._data_shape[0])
        # else:
        #     t = self._validtarget_ceilT - 1
        # 5
        # 1.2 => 2
        total_t, h, w, _ = self._data_shape
        framepixelcount = h * w
        available_h = h - self._h_validmax
        if available_h < self._patch_size:
            available_h = 0
        available_w = w - self._w_validmax
        if available_w < self._patch_size:
            available_w = 0

        targetpixels = np.array(
            [available_h * available_w]
            + [framepixelcount] * (total_t - self._validtarget_ceilT)
        )
        t_probab = targetpixels / np.sum(targetpixels)
        t = np.random.choice(
            np.arange(self._validtarget_ceilT - 1, total_t), p=t_probab
        )

        h, w = self.get_invalid_target_hw(t)
        index = self.idx_manager.idx_from_hwt(h, w, t)
        # print('Invalid', index, h,w,t)
        return index

    def get_valid_target_hw(self, t):
        """
        This is the opposite of get_invalid_target_hw. It returns a h,w which is valid for target.
        This is only valid for single frame setup.
        """
        if t == self._validtarget_ceilT - 1:
            h = np.random.randint(0, self._h_validmax - self._patch_size)
            w = np.random.randint(0, self._w_validmax - self._patch_size)
        else:
            h = np.random.randint(0, self._data_shape[1] - self._patch_size)
            w = np.random.randint(0, self._data_shape[2] - self._patch_size)
        return h, w

    def get_invalid_target_hw(self, t):
        """
        This is the opposite of get_valid_target_hw. It returns a h,w which is not valid for target.
        This is only valid for single frame setup.
        """
        if t == self._validtarget_ceilT - 1:
            h = np.random.randint(
                self._h_validmax, self._data_shape[1] - self._patch_size
            )
            w = np.random.randint(
                self._w_validmax, self._data_shape[2] - self._patch_size
            )
        else:
            h = np.random.randint(0, self._data_shape[1] - self._patch_size)
            w = np.random.randint(0, self._data_shape[2] - self._patch_size)
        return h, w

    def _get_tidx(self, index):
        if isinstance(index, int) or isinstance(index, np.int64):
            idx = index
        else:
            idx = index[0]
        return self.idx_manager.get_t(idx)

    def index_should_have_target(self, index):
        tidx = self._get_tidx(index)
        if tidx < self._validtarget_ceilT - 1:
            return True
        elif tidx > self._validtarget_ceilT - 1:
            return False
        else:
            h, w, _ = self.idx_manager.hwt_from_idx(index)
            return (
                h + self._patch_size < self._h_validmax
                and w + self._patch_size < self._w_validmax
            )

    @staticmethod
    def get_reduced_frame_size(data_shape_nhw, fraction):
        n, h, w = data_shape_nhw

        framepixelcount = h * w
        targetpixelcount = int(n * framepixelcount * fraction)

        # We are currently supporting this only when there is just one frame.
        # if np.ceil(pixelcount / framepixelcount) > 1:
        #     return None, None

        lastframepixelcount = targetpixelcount % framepixelcount
        assert data_shape_nhw[1] == data_shape_nhw[2]
        if lastframepixelcount > 0:
            new_size = int(np.sqrt(lastframepixelcount))
            return new_size, new_size
        else:
            assert (
                targetpixelcount / framepixelcount >= 1
            ), "This is not possible in euclidean space :D (so this is a bug)"
            return h, w



import numpy as np
import pytest

from careamics.config.tile_information import TileInformation
from careamics.dataset.tiling.lvae_tiled_patching import (
    compute_padding,
    compute_tile_grid_shape,
    compute_tile_info,
    compute_tile_info_legacy,
    extract_tiles,
    n_tiles_1d,
    total_n_tiles,
)
from careamics.lvae_training.dataset import TilingMode
from careamics.lvae_training.dataset.utils.index_manager import GridIndexManager
from careamics.prediction_utils.stitch_prediction import stitch_prediction


@pytest.mark.parametrize(
    "data_shape, tile_size, overlaps",
    [
        # 2D
        ((1, 3, 10, 9), (4, 4), (2, 2)),
        ((1, 3, 10, 9), (8, 8), (4, 4)),
        # 3D
        ((1, 3, 8, 16, 17), (4, 4, 4), (2, 2, 2)),
        ((1, 3, 8, 16, 17), (8, 8, 8), (4, 4, 4)),
    ],
)
def test_extract_tiles(data_shape, tile_size, overlaps):
    """Test extracted tiles are all the same size and can reconstruct the image."""

    arr = np.random.random_sample(data_shape).astype(np.float32)

    tile_data_generator = extract_tiles(
        arr=arr, tile_size=np.array(tile_size), overlaps=np.array(overlaps)
    )

    tiles = []
    tile_infos = []

    # Assemble all tiles and their respective coordinates
    for tile, tile_info in tile_data_generator:

        overlap_crop_coords = tile_info.overlap_crop_coords
        stitch_coords = tile_info.stitch_coords

        # add data to lists
        tiles.append(tile)
        tile_infos.append(tile_info)

        # check tile shape, ignore channel dimension
        assert tile.shape[1:] == tile_size
        assert len(overlap_crop_coords) == len(stitch_coords) == len(tile_size)

    # stitch_prediction returns list
    stitched_arr = stitch_prediction(tiles, tile_infos)[0]

    np.testing.assert_array_equal(arr, stitched_arr)


def test_compute_tile_info_legacy():
    """Test the legacy version of `compute_tile_info`."""
    data_shape = (1, 3, 10, 9)
    tile_size = (1, 3, 4, 4)
    overlaps = (0, 0, 2, 2)
    stitch_shape = tuple(ts - ov for ts, ov in zip(tile_size, overlaps))

    grid_index_manager = GridIndexManager(
        data_shape=data_shape,
        grid_shape=stitch_shape,
        patch_shape=tile_size,
        tiling_mode=TilingMode.PadBoundary,
    )

    # first example
    tile_info = compute_tile_info_legacy(grid_index_manager=grid_index_manager, index=0)
    assert tile_info == TileInformation(
        array_shape=tuple(data_shape[1:]),
        last_tile=False,
        overlap_crop_coords=((1, 3), (1, 3)),
        stitch_coords=((0, 2), (0, 2)),
        sample_id=0,
    )

    # second example
    tile_info = compute_tile_info_legacy(
        grid_index_manager=grid_index_manager, index=12
    )
    assert tile_info == TileInformation(
        array_shape=tuple(data_shape[1:]),
        last_tile=False,
        overlap_crop_coords=((1, 3), (1, 3)),
        stitch_coords=((4, 6), (4, 6)),
        sample_id=0,
    )

    # third example
    tile_info = compute_tile_info_legacy(
        grid_index_manager=grid_index_manager, index=14
    )
    assert tile_info == TileInformation(
        array_shape=tuple(data_shape[1:]),
        last_tile=False,
        overlap_crop_coords=((1, 3), (1, 2)),
        stitch_coords=((4, 6), (8, 9)),
        sample_id=0,
    )

    # fourth example
    tile_info = compute_tile_info_legacy(
        grid_index_manager=grid_index_manager, index=24
    )
    assert tile_info == TileInformation(
        array_shape=tuple(data_shape[1:]),
        last_tile=True,
        overlap_crop_coords=((1, 3), (1, 2)),
        stitch_coords=((8, 10), (8, 9)),
        sample_id=0,
    )


def test_compute_tile_info():
    """Test `compute_tile_info` for a selection of known results."""

    # TODO: improve this test ?

    data_shape = np.array([1, 3, 10, 9])
    tile_size = np.array([4, 4])
    overlaps = np.array([2, 2])

    # first example
    tile_info = compute_tile_info((0, 0), data_shape[1:], tile_size, overlaps)
    assert tile_info == TileInformation(
        array_shape=tuple(data_shape[1:]),
        last_tile=False,
        overlap_crop_coords=((1, 3), (1, 3)),
        stitch_coords=((0, 2), (0, 2)),
        sample_id=0,
    )

    # second example
    tile_info = compute_tile_info((2, 2), data_shape[1:], tile_size, overlaps)
    assert tile_info == TileInformation(
        array_shape=tuple(data_shape[1:]),
        last_tile=False,
        overlap_crop_coords=((1, 3), (1, 3)),
        stitch_coords=((4, 6), (4, 6)),
        sample_id=0,
    )

    # third example
    tile_info = compute_tile_info((2, 4), data_shape[1:], tile_size, overlaps)
    assert tile_info == TileInformation(
        array_shape=tuple(data_shape[1:]),
        last_tile=False,
        overlap_crop_coords=((1, 3), (1, 2)),
        stitch_coords=((4, 6), (8, 9)),
        sample_id=0,
    )

    # fourth example
    tile_info = compute_tile_info((4, 4), data_shape[1:], tile_size, overlaps)
    assert tile_info == TileInformation(
        array_shape=tuple(data_shape[1:]),
        last_tile=True,
        overlap_crop_coords=((1, 3), (1, 2)),
        stitch_coords=((8, 10), (8, 9)),
        sample_id=0,
    )


@pytest.mark.parametrize(
    "data_shape, tile_size, overlaps",
    [
        # 2D
        ((1, 3, 10, 9), (4, 4), (2, 2)),
        ((1, 3, 10, 9), (8, 8), (4, 4)),
        # 3D
        ((1, 3, 8, 16, 17), (4, 4, 4), (2, 2, 2)),
        ((1, 3, 8, 16, 17), (8, 8, 8), (4, 4, 4)),
    ],
)
def test_compute_padding(data_shape, tile_size, overlaps):

    padding = compute_padding(
        np.array(data_shape), np.array(tile_size), np.array(overlaps)
    )

    for axis, (before, after) in enumerate(padding):
        # padded array should be divisible by the stitch size
        stitch_size = tile_size[axis] - overlaps[axis]
        axis_size = data_shape[axis + 2]  # + 2 for sample and channel dims
        assert (before + axis_size + after) % stitch_size == 0

        assert before == overlaps[axis] // 2


@pytest.mark.parametrize(
    "axis_size, tile_size, overlap",
    [(9, 4, 2), (10, 8, 4), (17, 8, 4)],
)
def test_n_tiles_1d(axis_size, tile_size, overlap):
    """Test calculating the number of tiles in a specific dimension."""
    result = n_tiles_1d(axis_size, tile_size, overlap)
    assert result == int(np.ceil(axis_size / (tile_size - overlap)))


@pytest.mark.parametrize(
    "data_shape, tile_size, overlaps",
    [
        # 2D
        ((1, 3, 10, 9), (4, 4), (2, 2)),
        ((1, 3, 10, 9), (8, 8), (4, 4)),
        # 3D
        ((1, 3, 8, 16, 17), (4, 4, 4), (2, 2, 2)),
        ((1, 3, 8, 16, 17), (8, 8, 8), (4, 4, 4)),
    ],
)
def test_total_n_tiles(data_shape, tile_size, overlaps):
    """Test calculating the total number of tiles."""

    result = total_n_tiles(data_shape, tile_size, overlaps)
    n_tiles = 1
    for i in range(-1, -len(tile_size) - 1, -1):
        n_tiles = n_tiles * int(np.ceil(data_shape[i] / (tile_size[i] - overlaps[i])))

    assert result == n_tiles


@pytest.mark.parametrize(
    "data_shape, tile_size, overlaps",
    [
        # 2D
        ((1, 3, 10, 9), (4, 4), (2, 2)),
        ((1, 3, 10, 9), (8, 8), (4, 4)),
        # 3D
        ((1, 3, 8, 16, 17), (4, 4, 4), (2, 2, 2)),
        ((1, 3, 8, 16, 17), (8, 8, 8), (4, 4, 4)),
    ],
)
def test_compute_tile_grid_shape(data_shape, tile_size, overlaps):
    """Test computing tile grid shape."""

    result = compute_tile_grid_shape(data_shape, tile_size, overlaps)

    tile_grid_shape = tuple(
        int(np.ceil(data_shape[i] / (tile_size[i] - overlaps[i])))
        for i in range(-len(tile_size), 0, 1)
    )

    assert result == tile_grid_shape



"""
A place for Datasets and Dataloaders.
"""

from typing import Tuple, Union, Callable

import numpy as np
from skimage.transform import resize

from .config import DatasetConfig
from .multich_dataset import MultiChDloader


class LCMultiChDloader(MultiChDloader):
    def __init__(
        self,
        data_config: DatasetConfig,
        fpath: str,
        load_data_fn: Callable,
        val_fraction=None,
        test_fraction=None,
    ):
        self._padding_kwargs = (
            data_config.padding_kwargs  # mode=padding_mode, constant_values=constant_value
        )
        self._uncorrelated_channel_probab = data_config.uncorrelated_channel_probab

        super().__init__(
            data_config,
            fpath,
            load_data_fn=load_data_fn,
            val_fraction=val_fraction,
            test_fraction=test_fraction,
        )

        if data_config.overlapping_padding_kwargs is not None:
            assert (
                self._padding_kwargs == data_config.overlapping_padding_kwargs
            ), "During evaluation, overlapping_padding_kwargs should be same as padding_args. \
                It should be so since we just use overlapping_padding_kwargs when it is not None"

        else:
            self._overlapping_padding_kwargs = data_config.padding_kwargs

        self.multiscale_lowres_count = data_config.multiscale_lowres_count
        assert self.multiscale_lowres_count is not None
        self._scaled_data = [self._data]
        self._scaled_noise_data = [self._noise_data]

        assert (
            isinstance(self.multiscale_lowres_count, int)
            and self.multiscale_lowres_count >= 1
        )
        assert isinstance(self._padding_kwargs, dict)
        assert "mode" in self._padding_kwargs

        for _ in range(1, self.multiscale_lowres_count):
            shape = self._scaled_data[-1].shape
            assert len(shape) == 4
            new_shape = (shape[0], shape[1] // 2, shape[2] // 2, shape[3])
            ds_data = resize(
                self._scaled_data[-1].astype(np.float32), new_shape
            ).astype(self._scaled_data[-1].dtype)
            # NOTE: These asserts are important. the resize method expects np.float32. otherwise, one gets weird results.
            assert (
                ds_data.max() / self._scaled_data[-1].max() < 5
            ), "Downsampled image should not have very different values"
            assert (
                ds_data.max() / self._scaled_data[-1].max() > 0.2
            ), "Downsampled image should not have very different values"

            self._scaled_data.append(ds_data)
            # do the same for noise
            if self._noise_data is not None:
                noise_data = resize(self._scaled_noise_data[-1], new_shape)
                self._scaled_noise_data.append(noise_data)

    def reduce_data(
        self, t_list=None, h_start=None, h_end=None, w_start=None, w_end=None
    ):
        assert t_list is not None
        assert h_start is None
        assert h_end is None
        assert w_start is None
        assert w_end is None

        self._data = self._data[t_list].copy()
        self._scaled_data = [
            self._scaled_data[i][t_list].copy() for i in range(len(self._scaled_data))
        ]

        if self._noise_data is not None:
            self._noise_data = self._noise_data[t_list].copy()
            self._scaled_noise_data = [
                self._scaled_noise_data[i][t_list].copy()
                for i in range(len(self._scaled_noise_data))
            ]

        self.N = len(t_list)
        self.set_img_sz(self._img_sz, self._grid_sz)
        print(
            f"[{self.__class__.__name__}] Data reduced. New data shape: {self._data.shape}"
        )

    def _init_msg(self):
        msg = super()._init_msg()
        msg += f" Pad:{self._padding_kwargs}"
        if self._uncorrelated_channels:
            msg += f" UncorrChProbab:{self._uncorrelated_channel_probab}"
        return msg

    def _load_scaled_img(
        self, scaled_index, index: Union[int, Tuple[int, int]]
    ) -> Tuple[np.ndarray, np.ndarray]:
        if isinstance(index, int):
            idx = index
        else:
            idx, _ = index

        # tidx = self.idx_manager.get_t(idx)
        patch_loc_list = self.idx_manager.get_patch_location_from_dataset_idx(idx)
        nidx = patch_loc_list[0]

        imgs = self._scaled_data[scaled_index][nidx]
        imgs = tuple([imgs[None, ..., i] for i in range(imgs.shape[-1])])
        if self._noise_data is not None:
            noisedata = self._scaled_noise_data[scaled_index][nidx]
            noise = tuple([noisedata[None, ..., i] for i in range(noisedata.shape[-1])])
            factor = np.sqrt(2) if self._input_is_sum else 1.0
            imgs = tuple([img + noise[0] * factor for img in imgs])
        return imgs

    def _crop_img(self, img: np.ndarray, patch_start_loc: Tuple):
        """
        Here, h_start, w_start could be negative. That simply means we need to pick the content from 0. So,
        the cropped image will be smaller than self._img_sz * self._img_sz
        """
        max_len_vals = list(self.idx_manager.data_shape[1:-1])
        max_len_vals[-2:] = img.shape[-2:]
        return self._crop_img_with_padding(
            img, patch_start_loc, max_len_vals=max_len_vals
        )

    def _get_img(self, index: int):
        """
        Returns the primary patch along with low resolution patches centered on the primary patch.
        """
        # Noise_tuples is populated when there is synthetic noise in training
        # Should have similar type of noise with the noise model
        # Starting with microsplit, dump the noise, use it instead as an augmentation if nessesary
        img_tuples, noise_tuples = self._load_img(index)
        assert self._img_sz is not None
        h, w = img_tuples[0].shape[-2:]
        if self._enable_random_cropping:
            patch_start_loc = self._get_random_hw(h, w)
            if self._5Ddata:
                patch_start_loc = (
                    np.random.choice(img_tuples[0].shape[-3] - self._depth3D),
                ) + patch_start_loc
        else:
            patch_start_loc = self._get_deterministic_loc(index)

        # LC logic is located here, the function crops the image of the highest resolution
        cropped_img_tuples = [
            self._crop_flip_img(img, patch_start_loc, False, False)
            for img in img_tuples
        ]
        cropped_noise_tuples = [
            self._crop_flip_img(noise, patch_start_loc, False, False)
            for noise in noise_tuples
        ]
        patch_start_loc = list(patch_start_loc)
        h_start, w_start = patch_start_loc[-2], patch_start_loc[-1]
        h_center = h_start + self._img_sz // 2
        w_center = w_start + self._img_sz // 2
        allres_versions = {
            i: [cropped_img_tuples[i]] for i in range(len(cropped_img_tuples))
        }
        for scale_idx in range(1, self.multiscale_lowres_count):
            # Returning the image of the lower resolution
            scaled_img_tuples = self._load_scaled_img(scale_idx, index)

            h_center = h_center // 2
            w_center = w_center // 2

            h_start = h_center - self._img_sz // 2
            w_start = w_center - self._img_sz // 2
            patch_start_loc[-2:] = [h_start, w_start]
            scaled_cropped_img_tuples = [
                self._crop_flip_img(img, patch_start_loc, False, False)
                for img in scaled_img_tuples
            ]
            for ch_idx in range(len(img_tuples)):
                allres_versions[ch_idx].append(scaled_cropped_img_tuples[ch_idx])

        output_img_tuples = tuple(
            [
                np.concatenate(allres_versions[ch_idx])
                for ch_idx in range(len(img_tuples))
            ]
        )
        return output_img_tuples, cropped_noise_tuples

    def __getitem__(self, index: Union[int, Tuple[int, int]]):
        img_tuples, noise_tuples = self._get_img(index)
        if self._uncorrelated_channels:
            assert (
                self._input_idx is None
            ), "Uncorrelated channels is not implemented when there is a separate input channel."
            if np.random.rand() < self._uncorrelated_channel_probab:
                img_tuples_new = [None] * len(img_tuples)
                img_tuples_new[0] = img_tuples[0]
                for i in range(1, len(img_tuples)):
                    new_index = np.random.randint(len(self))
                    img_tuples_tmp, _ = self._get_img(new_index)
                    img_tuples_new[i] = img_tuples_tmp[i]
                img_tuples = img_tuples_new

        if self._is_train:
            if self._empty_patch_replacement_enabled:
                if np.random.rand() < self._empty_patch_replacement_probab:
                    img_tuples = self.replace_with_empty_patch(img_tuples)

        if self._enable_rotation:
            img_tuples, noise_tuples = self._rotate(img_tuples, noise_tuples)

        # add noise to input, if noise is present combine it with the image
        # factor is for the compute input not to have too much noise because the average of two gaussians
        if len(noise_tuples) > 0:
            factor = np.sqrt(2) if self._input_is_sum else 1.0
            input_tuples = []
            for x in img_tuples:
                x = (
                    x.copy()
                )  # to avoid changing the original image since it is later used for target
                # NOTE: other LC levels already have noise added. So, we just need to add noise to the highest resolution.
                x[0] = x[0] + noise_tuples[0] * factor
                input_tuples.append(x)
        else:
            input_tuples = img_tuples

        # Compute the input by sum / average the channels
        # Alpha is an amount of weight which is applied to the channels when combining them
        # How to sample alpha is still under research
        inp, alpha = self._compute_input(input_tuples)
        target_tuples = [img[:1] for img in img_tuples]
        # add noise to target.
        if len(noise_tuples) >= 1:
            target_tuples = [
                x + noise for x, noise in zip(target_tuples, noise_tuples[1:])
            ]

        target = self._compute_target(target_tuples, alpha)

        norm_target = self.normalize_target(target)

        output = [inp, norm_target]

        if self._return_alpha:
            output.append(alpha)

        if isinstance(index, int):
            return tuple(output)

        _, grid_size = index
        output.append(grid_size)
        return tuple(output)



from pathlib import Path
from unittest.mock import Mock

from careamics.config import InferenceConfig
from careamics.dataset import IterablePredDataset, IterableTiledPredDataset
from careamics.lightning.callbacks.prediction_writer_callback.file_path_utils import (
    create_write_file_path,
    get_sample_file_path,
)


def test_get_sample_file_path_tiled_ds():

    # Create DS with mock InferenceConfig
    src_files = [f"{i}.ext" for i in range(2)]
    pred_config = Mock(spec=InferenceConfig)
    # attrs used in DS initialization
    pred_config.axes = Mock()
    pred_config.tile_size = Mock()
    pred_config.tile_overlap = Mock()
    pred_config.image_means = [Mock()]
    pred_config.image_stds = [Mock()]
    ds = IterableTiledPredDataset(pred_config, src_files=src_files)

    for i in range(2):
        file_path = get_sample_file_path(ds, sample_id=i)
        assert file_path == f"{i}.ext"


def test_get_sample_file_path_untiled_ds():

    # Create DS with mock InferenceConfig
    src_files = [f"{i}.ext" for i in range(2)]
    pred_config = Mock(spec=InferenceConfig)
    # attrs used in DS initialization
    pred_config.axes = Mock()
    pred_config.image_means = [Mock()]
    pred_config.image_stds = [Mock()]
    ds = IterablePredDataset(pred_config, src_files=src_files)

    for i in range(2):
        file_path = get_sample_file_path(ds, sample_id=i)
        assert file_path == f"{i}.ext"


def test_create_write_file_path():
    dirpath = Path("output_directory")
    file_path = Path("input_directory/file_name.in_ext")
    write_extension = ".out_ext"

    write_file_path = create_write_file_path(
        dirpath=dirpath, file_path=file_path, write_extension=write_extension
    )
    assert write_file_path == Path("output_directory/file_name.out_ext")



"""Tiled patching utilities."""

import itertools
from typing import Generator, List, Tuple, Union

import numpy as np

from careamics.config.tile_information import TileInformation


def _compute_crop_and_stitch_coords_1d(
    axis_size: int, tile_size: int, overlap: int
) -> Tuple[List[Tuple[int, int]], List[Tuple[int, int]], List[Tuple[int, int]]]:
    """
    Compute the coordinates of each tile along an axis, given the overlap.

    Parameters
    ----------
    axis_size : int
        Length of the axis.
    tile_size : int
        Size of the tile for the given axis.
    overlap : int
        Size of the overlap for the given axis.

    Returns
    -------
    Tuple[Tuple[int, ...], ...]
        Tuple of all coordinates for given axis.
    """
    # Compute the step between tiles
    step = tile_size - overlap
    crop_coords = []
    stitch_coords = []
    overlap_crop_coords = []

    # Iterate over the axis with step
    for i in range(0, max(1, axis_size - overlap), step):
        # Check if the tile fits within the axis
        if i + tile_size <= axis_size:
            # Add the coordinates to crop one tile
            crop_coords.append((i, i + tile_size))

            # Add the pixel coordinates of the cropped tile in the original image space
            stitch_coords.append(
                (
                    i + overlap // 2 if i > 0 else 0,
                    (
                        i + tile_size - overlap // 2
                        if crop_coords[-1][1] < axis_size
                        else axis_size
                    ),
                )
            )

            # Add the coordinates to crop the overlap from the prediction.
            overlap_crop_coords.append(
                (
                    overlap // 2 if i > 0 else 0,
                    (
                        tile_size - overlap // 2
                        if crop_coords[-1][1] < axis_size
                        else tile_size
                    ),
                )
            )

        # If the tile does not fit within the axis, perform the abovementioned
        # operations starting from the end of the axis
        else:
            # if (axis_size - tile_size, axis_size) not in crop_coords:
            crop_coords.append((max(0, axis_size - tile_size), axis_size))
            last_tile_end_coord = stitch_coords[-1][1] if stitch_coords else 1
            stitch_coords.append((last_tile_end_coord, axis_size))
            overlap_crop_coords.append(
                (tile_size - (axis_size - last_tile_end_coord), tile_size)
            )
            break
    return crop_coords, stitch_coords, overlap_crop_coords


def extract_tiles(
    arr: np.ndarray,
    tile_size: Union[List[int], Tuple[int, ...]],
    overlaps: Union[List[int], Tuple[int, ...]],
) -> Generator[Tuple[np.ndarray, TileInformation], None, None]:
    """Generate tiles from the input array with specified overlap.

    The tiles cover the whole array. The method returns a generator that yields
    tuples of array and tile information, the latter includes whether
    the tile is the last one, the coordinates of the overlap crop, and the coordinates
    of the stitched tile.

    Input array should have shape SC(Z)YX, while the returned tiles have shape C(Z)YX,
    where C can be a singleton.

    Parameters
    ----------
    arr : np.ndarray
        Array of shape (S, C, (Z), Y, X).
    tile_size : Union[List[int], Tuple[int]]
        Tile sizes in each dimension, of length 2 or 3.
    overlaps : Union[List[int], Tuple[int]]
        Overlap values in each dimension, of length 2 or 3.

    Yields
    ------
    Generator[Tuple[np.ndarray, TileInformation], None, None]
        Tile generator, yields the tile and additional information.
    """
    # Iterate over num samples (S)
    for sample_idx in range(arr.shape[0]):
        sample: np.ndarray = arr[sample_idx, ...]

        # Create a list of coordinates for cropping and stitching all axes.
        # [crop coordinates, stitching coordinates, overlap crop coordinates]
        # For axis of size 35 and patch size of 32 compute_crop_and_stitch_coords_1d
        # will output ([(0, 32), (3, 35)], [(0, 20), (20, 35)], [(0, 20), (17, 32)])
        crop_and_stitch_coords_list = [
            _compute_crop_and_stitch_coords_1d(
                sample.shape[i + 1], tile_size[i], overlaps[i]
            )
            for i in range(len(tile_size))
        ]

        # Rearrange crop coordinates from a list of coordinate pairs per axis to a list
        # grouped by type.
        all_crop_coords, all_stitch_coords, all_overlap_crop_coords = zip(
            *crop_and_stitch_coords_list
        )

        # Maximum tile index
        max_tile_idx = np.prod([len(axis) for axis in all_crop_coords]) - 1

        # Iterate over generated coordinate pairs:
        for tile_idx, (crop_coords, stitch_coords, overlap_crop_coords) in enumerate(
            zip(
                itertools.product(*all_crop_coords),
                itertools.product(*all_stitch_coords),
                itertools.product(*all_overlap_crop_coords),
            )
        ):
            # Extract tile from the sample
            tile: np.ndarray = sample[
                (..., *[slice(c[0], c[1]) for c in list(crop_coords)])  # type: ignore
            ]

            # Check if we are at the end of the sample by computing the length of the
            # array that contains all the tiles
            if tile_idx == max_tile_idx:
                last_tile = True
            else:
                last_tile = False

            # create tile information
            tile_info = TileInformation(
                array_shape=sample.shape,
                last_tile=last_tile,
                overlap_crop_coords=overlap_crop_coords,
                stitch_coords=stitch_coords,
                sample_id=sample_idx,
            )

            yield tile, tile_info



"""Transforms that are used to augment the data."""

__all__ = [
    "get_all_transforms",
    "N2VManipulate",
    "XYFlip",
    "XYRandomRotate90",
    "ImageRestorationTTA",
    "Denormalize",
    "Normalize",
    "Compose",
]


from .compose import Compose, get_all_transforms
from .n2v_manipulate import N2VManipulate
from .normalize import Denormalize, Normalize
from .tta import ImageRestorationTTA
from .xy_flip import XYFlip
from .xy_random_rotate90 import XYRandomRotate90



from dataclasses import dataclass

import numpy as np

from careamics.lvae_training.dataset.types import TilingMode


@dataclass
class GridIndexManager:
    data_shape: tuple
    grid_shape: tuple
    patch_shape: tuple
    tiling_mode: TilingMode

    # Patch is centered on index in the grid, grid size not used in training,
    # used only during val / test, grid size controls the overlap of the patches
    # in training you only get random patches every time
    # For borders - just cropped the data, so it perfectly divisible

    def __post_init__(self):
        assert len(self.data_shape) == len(
            self.grid_shape
        ), f"Data shape:{self.data_shape} and grid size:{self.grid_shape} must have the same dimension"
        assert len(self.data_shape) == len(
            self.patch_shape
        ), f"Data shape:{self.data_shape} and patch shape:{self.patch_shape} must have the same dimension"
        innerpad = np.array(self.patch_shape) - np.array(self.grid_shape)
        for dim, pad in enumerate(innerpad):
            if pad < 0:
                raise ValueError(
                    f"Patch shape:{self.patch_shape} must be greater than or equal to grid shape:{self.grid_shape} in dimension {dim}"
                )
            if pad % 2 != 0:
                raise ValueError(
                    f"Patch shape:{self.patch_shape} must have even padding in dimension {dim}"
                )

    def patch_offset(self):
        return (np.array(self.patch_shape) - np.array(self.grid_shape)) // 2

    def get_individual_dim_grid_count(self, dim: int):
        """
        Returns the number of the grid in the specified dimension, ignoring all other dimensions.
        """
        assert dim < len(
            self.data_shape
        ), f"Dimension {dim} is out of bounds for data shape {self.data_shape}"
        assert dim >= 0, "Dimension must be greater than or equal to 0"

        if self.grid_shape[dim] == 1 and self.patch_shape[dim] == 1:
            return self.data_shape[dim]
        elif self.tiling_mode == TilingMode.PadBoundary:
            return int(np.ceil(self.data_shape[dim] / self.grid_shape[dim]))
        elif self.tiling_mode == TilingMode.ShiftBoundary:
            excess_size = self.patch_shape[dim] - self.grid_shape[dim]
            return int(
                np.ceil((self.data_shape[dim] - excess_size) / self.grid_shape[dim])
            )
        else:
            excess_size = self.patch_shape[dim] - self.grid_shape[dim]
            return int(
                np.floor((self.data_shape[dim] - excess_size) / self.grid_shape[dim])
            )

    def total_grid_count(self):
        """
        Returns the total number of grids in the dataset.
        """
        return self.grid_count(0) * self.get_individual_dim_grid_count(0)

    def grid_count(self, dim: int):
        """
        Returns the total number of grids for one value in the specified dimension.
        """
        assert dim < len(
            self.data_shape
        ), f"Dimension {dim} is out of bounds for data shape {self.data_shape}"
        assert dim >= 0, "Dimension must be greater than or equal to 0"
        if dim == len(self.data_shape) - 1:
            return 1

        return self.get_individual_dim_grid_count(dim + 1) * self.grid_count(dim + 1)

    def get_grid_index(self, dim: int, coordinate: int):
        """
        Returns the index of the grid in the specified dimension.
        """
        assert dim < len(
            self.data_shape
        ), f"Dimension {dim} is out of bounds for data shape {self.data_shape}"
        assert dim >= 0, "Dimension must be greater than or equal to 0"
        assert (
            coordinate < self.data_shape[dim]
        ), f"Coordinate {coordinate} is out of bounds for data shape {self.data_shape}"

        if self.grid_shape[dim] == 1 and self.patch_shape[dim] == 1:
            return coordinate
        elif self.tiling_mode == TilingMode.PadBoundary:  # self.trim_boundary is False:
            return np.floor(coordinate / self.grid_shape[dim])
        elif self.tiling_mode == TilingMode.TrimBoundary:
            excess_size = (self.patch_shape[dim] - self.grid_shape[dim]) // 2
            # can be <0 if coordinate is in [0,grid_shape[dim]]
            return max(0, np.floor((coordinate - excess_size) / self.grid_shape[dim]))
        elif self.tiling_mode == TilingMode.ShiftBoundary:
            excess_size = (self.patch_shape[dim] - self.grid_shape[dim]) // 2
            if coordinate + self.grid_shape[dim] + excess_size == self.data_shape[dim]:
                return self.get_individual_dim_grid_count(dim) - 1
            else:
                # can be <0 if coordinate is in [0,grid_shape[dim]]
                return max(
                    0, np.floor((coordinate - excess_size) / self.grid_shape[dim])
                )

        else:
            raise ValueError(f"Unsupported tiling mode {self.tiling_mode}")

    def dataset_idx_from_grid_idx(self, grid_idx: tuple):
        """
        Returns the index of the grid in the dataset.
        """
        assert len(grid_idx) == len(
            self.data_shape
        ), f"Dimension indices {grid_idx} must have the same dimension as data shape {self.data_shape}"
        index = 0
        for dim in range(len(grid_idx)):
            index += grid_idx[dim] * self.grid_count(dim)
        return index

    def get_patch_location_from_dataset_idx(self, dataset_idx: int):
        """
        Returns the patch location of the grid in the dataset.
        """
        grid_location = self.get_location_from_dataset_idx(dataset_idx)
        offset = self.patch_offset()
        return tuple(np.array(grid_location) - np.array(offset))

    def get_dataset_idx_from_grid_location(self, location: tuple):
        assert len(location) == len(
            self.data_shape
        ), f"Location {location} must have the same dimension as data shape {self.data_shape}"
        grid_idx = [
            self.get_grid_index(dim, location[dim]) for dim in range(len(location))
        ]
        return self.dataset_idx_from_grid_idx(tuple(grid_idx))

    def get_gridstart_location_from_dim_index(self, dim: int, dim_index: int):
        """
        Returns the grid-start coordinate of the grid in the specified dimension.
        """
        assert dim < len(
            self.data_shape
        ), f"Dimension {dim} is out of bounds for data shape {self.data_shape}"
        assert dim >= 0, "Dimension must be greater than or equal to 0"
        # assert dim_index < self.get_individual_dim_grid_count(
        #     dim
        # ), f"Dimension index {dim_index} is out of bounds for data shape {self.data_shape}"
        # TODO comented out this shit cuz I have no interest to dig why it's failing at this point !
        if self.grid_shape[dim] == 1 and self.patch_shape[dim] == 1:
            return dim_index
        elif self.tiling_mode == TilingMode.PadBoundary:
            return dim_index * self.grid_shape[dim]
        elif self.tiling_mode == TilingMode.TrimBoundary:
            excess_size = (self.patch_shape[dim] - self.grid_shape[dim]) // 2
            return dim_index * self.grid_shape[dim] + excess_size
        elif self.tiling_mode == TilingMode.ShiftBoundary:
            excess_size = (self.patch_shape[dim] - self.grid_shape[dim]) // 2
            if dim_index < self.get_individual_dim_grid_count(dim) - 1:
                return dim_index * self.grid_shape[dim] + excess_size
            else:
                # on boundary. grid should be placed such that the patch covers the entire data.
                return self.data_shape[dim] - self.grid_shape[dim] - excess_size
        else:
            raise ValueError(f"Unsupported tiling mode {self.tiling_mode}")

    def get_location_from_dataset_idx(self, dataset_idx: int):
        """
        Returns the start location of the grid in the dataset.
        """
        grid_idx = []
        for dim in range(len(self.data_shape)):
            grid_idx.append(dataset_idx // self.grid_count(dim))
            dataset_idx = dataset_idx % self.grid_count(dim)
        location = [
            self.get_gridstart_location_from_dim_index(dim, grid_idx[dim])
            for dim in range(len(self.data_shape))
        ]
        return tuple(location)

    def on_boundary(self, dataset_idx: int, dim: int, only_end: bool = False):
        """
        Returns True if the grid is on the boundary in the specified dimension.
        """
        assert dim < len(
            self.data_shape
        ), f"Dimension {dim} is out of bounds for data shape {self.data_shape}"
        assert dim >= 0, "Dimension must be greater than or equal to 0"

        if dim > 0:
            dataset_idx = dataset_idx % self.grid_count(dim - 1)

        dim_index = dataset_idx // self.grid_count(dim)
        if only_end:
            return dim_index == self.get_individual_dim_grid_count(dim) - 1

        return (
            dim_index == 0 or dim_index == self.get_individual_dim_grid_count(dim) - 1
        )

    def next_grid_along_dim(self, dataset_idx: int, dim: int):
        """
        Returns the index of the grid in the specified dimension in the specified direction.
        """
        assert dim < len(
            self.data_shape
        ), f"Dimension {dim} is out of bounds for data shape {self.data_shape}"
        assert dim >= 0, "Dimension must be greater than or equal to 0"
        new_idx = dataset_idx + self.grid_count(dim)
        if new_idx >= self.total_grid_count():
            return None
        return new_idx

    def prev_grid_along_dim(self, dataset_idx: int, dim: int):
        """
        Returns the index of the grid in the specified dimension in the specified direction.
        """
        assert dim < len(
            self.data_shape
        ), f"Dimension {dim} is out of bounds for data shape {self.data_shape}"
        assert dim >= 0, "Dimension must be greater than or equal to 0"
        new_idx = dataset_idx - self.grid_count(dim)
        if new_idx < 0:
            return None
